# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
import grpc

import qrl.generated.qrl_pb2 as qrl__pb2


class PublicAPIStub(object):
  """//////////////////////////
  //////////////////////////
  //////////////////////////
  ////     API       ///////
  //////////////////////////
  //////////////////////////
  //////////////////////////

  This service describes the Public API used by clients (wallet/cli/etc)
  """

  def __init__(self, channel):
    """Constructor.

    Args:
      channel: A grpc.Channel.
    """
    self.GetNodeState = channel.unary_unary(
        '/qrl.PublicAPI/GetNodeState',
        request_serializer=qrl__pb2.GetNodeStateReq.SerializeToString,
        response_deserializer=qrl__pb2.GetNodeStateResp.FromString,
        )
    self.GetKnownPeers = channel.unary_unary(
        '/qrl.PublicAPI/GetKnownPeers',
        request_serializer=qrl__pb2.GetKnownPeersReq.SerializeToString,
        response_deserializer=qrl__pb2.GetKnownPeersResp.FromString,
        )
    self.GetPeersStat = channel.unary_unary(
        '/qrl.PublicAPI/GetPeersStat',
        request_serializer=qrl__pb2.GetPeersStatReq.SerializeToString,
        response_deserializer=qrl__pb2.GetPeersStatResp.FromString,
        )
    self.GetStats = channel.unary_unary(
        '/qrl.PublicAPI/GetStats',
        request_serializer=qrl__pb2.GetStatsReq.SerializeToString,
        response_deserializer=qrl__pb2.GetStatsResp.FromString,
        )
    self.GetAddressState = channel.unary_unary(
        '/qrl.PublicAPI/GetAddressState',
        request_serializer=qrl__pb2.GetAddressStateReq.SerializeToString,
        response_deserializer=qrl__pb2.GetAddressStateResp.FromString,
        )
    self.GetObject = channel.unary_unary(
        '/qrl.PublicAPI/GetObject',
        request_serializer=qrl__pb2.GetObjectReq.SerializeToString,
        response_deserializer=qrl__pb2.GetObjectResp.FromString,
        )
    self.GetLatestData = channel.unary_unary(
        '/qrl.PublicAPI/GetLatestData',
        request_serializer=qrl__pb2.GetLatestDataReq.SerializeToString,
        response_deserializer=qrl__pb2.GetLatestDataResp.FromString,
        )
    self.PushTransaction = channel.unary_unary(
        '/qrl.PublicAPI/PushTransaction',
        request_serializer=qrl__pb2.PushTransactionReq.SerializeToString,
        response_deserializer=qrl__pb2.PushTransactionResp.FromString,
        )
    self.TransferCoins = channel.unary_unary(
        '/qrl.PublicAPI/TransferCoins',
        request_serializer=qrl__pb2.TransferCoinsReq.SerializeToString,
        response_deserializer=qrl__pb2.TransferCoinsResp.FromString,
        )
    self.ParseAddress = channel.unary_unary(
        '/qrl.PublicAPI/ParseAddress',
        request_serializer=qrl__pb2.ParseAddressReq.SerializeToString,
        response_deserializer=qrl__pb2.ParseAddressResp.FromString,
        )
    self.GetAddressFromPK = channel.unary_unary(
        '/qrl.PublicAPI/GetAddressFromPK',
        request_serializer=qrl__pb2.GetAddressFromPKReq.SerializeToString,
        response_deserializer=qrl__pb2.GetAddressFromPKResp.FromString,
        )
    self.GetMessageTxn = channel.unary_unary(
        '/qrl.PublicAPI/GetMessageTxn',
        request_serializer=qrl__pb2.MessageTxnReq.SerializeToString,
        response_deserializer=qrl__pb2.TransferCoinsResp.FromString,
        )
    self.GetTokenTxn = channel.unary_unary(
        '/qrl.PublicAPI/GetTokenTxn',
        request_serializer=qrl__pb2.TokenTxnReq.SerializeToString,
        response_deserializer=qrl__pb2.TransferCoinsResp.FromString,
        )
    self.GetTransferTokenTxn = channel.unary_unary(
        '/qrl.PublicAPI/GetTransferTokenTxn',
        request_serializer=qrl__pb2.TransferTokenTxnReq.SerializeToString,
        response_deserializer=qrl__pb2.TransferCoinsResp.FromString,
        )
    self.GetSlaveTxn = channel.unary_unary(
        '/qrl.PublicAPI/GetSlaveTxn',
        request_serializer=qrl__pb2.SlaveTxnReq.SerializeToString,
        response_deserializer=qrl__pb2.TransferCoinsResp.FromString,
        )
    self.GetTransactionsByAddress = channel.unary_unary(
        '/qrl.PublicAPI/GetTransactionsByAddress',
        request_serializer=qrl__pb2.GetTransactionsByAddressReq.SerializeToString,
        response_deserializer=qrl__pb2.GetTransactionsByAddressResp.FromString,
        )
    self.GetTransaction = channel.unary_unary(
        '/qrl.PublicAPI/GetTransaction',
        request_serializer=qrl__pb2.GetTransactionReq.SerializeToString,
        response_deserializer=qrl__pb2.GetTransactionResp.FromString,
        )
    self.GetBalance = channel.unary_unary(
        '/qrl.PublicAPI/GetBalance',
        request_serializer=qrl__pb2.GetBalanceReq.SerializeToString,
        response_deserializer=qrl__pb2.GetBalanceResp.FromString,
        )
    self.GetOTS = channel.unary_unary(
        '/qrl.PublicAPI/GetOTS',
        request_serializer=qrl__pb2.GetOTSReq.SerializeToString,
        response_deserializer=qrl__pb2.GetOTSResp.FromString,
        )
    self.GetHeight = channel.unary_unary(
        '/qrl.PublicAPI/GetHeight',
        request_serializer=qrl__pb2.GetHeightReq.SerializeToString,
        response_deserializer=qrl__pb2.GetHeightResp.FromString,
        )
    self.GetBlock = channel.unary_unary(
        '/qrl.PublicAPI/GetBlock',
        request_serializer=qrl__pb2.GetBlockReq.SerializeToString,
        response_deserializer=qrl__pb2.GetBlockResp.FromString,
        )
    self.GetBlockByNumber = channel.unary_unary(
        '/qrl.PublicAPI/GetBlockByNumber',
        request_serializer=qrl__pb2.GetBlockByNumberReq.SerializeToString,
        response_deserializer=qrl__pb2.GetBlockByNumberResp.FromString,
        )


class PublicAPIServicer(object):
  """//////////////////////////
  //////////////////////////
  //////////////////////////
  ////     API       ///////
  //////////////////////////
  //////////////////////////
  //////////////////////////

  This service describes the Public API used by clients (wallet/cli/etc)
  """

  def GetNodeState(self, request, context):
    # missing associated documentation comment in .proto file
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetKnownPeers(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetPeersStat(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetStats(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetAddressState(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetObject(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetLatestData(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def PushTransaction(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def TransferCoins(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def ParseAddress(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetAddressFromPK(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetMessageTxn(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetTokenTxn(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetTransferTokenTxn(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetSlaveTxn(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetTransactionsByAddress(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetTransaction(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetBalance(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetOTS(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetHeight(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetBlock(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetBlockByNumber(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')


def add_PublicAPIServicer_to_server(servicer, server):
  rpc_method_handlers = {
      'GetNodeState': grpc.unary_unary_rpc_method_handler(
          servicer.GetNodeState,
          request_deserializer=qrl__pb2.GetNodeStateReq.FromString,
          response_serializer=qrl__pb2.GetNodeStateResp.SerializeToString,
      ),
      'GetKnownPeers': grpc.unary_unary_rpc_method_handler(
          servicer.GetKnownPeers,
          request_deserializer=qrl__pb2.GetKnownPeersReq.FromString,
          response_serializer=qrl__pb2.GetKnownPeersResp.SerializeToString,
      ),
      'GetPeersStat': grpc.unary_unary_rpc_method_handler(
          servicer.GetPeersStat,
          request_deserializer=qrl__pb2.GetPeersStatReq.FromString,
          response_serializer=qrl__pb2.GetPeersStatResp.SerializeToString,
      ),
      'GetStats': grpc.unary_unary_rpc_method_handler(
          servicer.GetStats,
          request_deserializer=qrl__pb2.GetStatsReq.FromString,
          response_serializer=qrl__pb2.GetStatsResp.SerializeToString,
      ),
      'GetAddressState': grpc.unary_unary_rpc_method_handler(
          servicer.GetAddressState,
          request_deserializer=qrl__pb2.GetAddressStateReq.FromString,
          response_serializer=qrl__pb2.GetAddressStateResp.SerializeToString,
      ),
      'GetObject': grpc.unary_unary_rpc_method_handler(
          servicer.GetObject,
          request_deserializer=qrl__pb2.GetObjectReq.FromString,
          response_serializer=qrl__pb2.GetObjectResp.SerializeToString,
      ),
      'GetLatestData': grpc.unary_unary_rpc_method_handler(
          servicer.GetLatestData,
          request_deserializer=qrl__pb2.GetLatestDataReq.FromString,
          response_serializer=qrl__pb2.GetLatestDataResp.SerializeToString,
      ),
      'PushTransaction': grpc.unary_unary_rpc_method_handler(
          servicer.PushTransaction,
          request_deserializer=qrl__pb2.PushTransactionReq.FromString,
          response_serializer=qrl__pb2.PushTransactionResp.SerializeToString,
      ),
      'TransferCoins': grpc.unary_unary_rpc_method_handler(
          servicer.TransferCoins,
          request_deserializer=qrl__pb2.TransferCoinsReq.FromString,
          response_serializer=qrl__pb2.TransferCoinsResp.SerializeToString,
      ),
      'ParseAddress': grpc.unary_unary_rpc_method_handler(
          servicer.ParseAddress,
          request_deserializer=qrl__pb2.ParseAddressReq.FromString,
          response_serializer=qrl__pb2.ParseAddressResp.SerializeToString,
      ),
      'GetAddressFromPK': grpc.unary_unary_rpc_method_handler(
          servicer.GetAddressFromPK,
          request_deserializer=qrl__pb2.GetAddressFromPKReq.FromString,
          response_serializer=qrl__pb2.GetAddressFromPKResp.SerializeToString,
      ),
      'GetMessageTxn': grpc.unary_unary_rpc_method_handler(
          servicer.GetMessageTxn,
          request_deserializer=qrl__pb2.MessageTxnReq.FromString,
          response_serializer=qrl__pb2.TransferCoinsResp.SerializeToString,
      ),
      'GetTokenTxn': grpc.unary_unary_rpc_method_handler(
          servicer.GetTokenTxn,
          request_deserializer=qrl__pb2.TokenTxnReq.FromString,
          response_serializer=qrl__pb2.TransferCoinsResp.SerializeToString,
      ),
      'GetTransferTokenTxn': grpc.unary_unary_rpc_method_handler(
          servicer.GetTransferTokenTxn,
          request_deserializer=qrl__pb2.TransferTokenTxnReq.FromString,
          response_serializer=qrl__pb2.TransferCoinsResp.SerializeToString,
      ),
      'GetSlaveTxn': grpc.unary_unary_rpc_method_handler(
          servicer.GetSlaveTxn,
          request_deserializer=qrl__pb2.SlaveTxnReq.FromString,
          response_serializer=qrl__pb2.TransferCoinsResp.SerializeToString,
      ),
      'GetTransactionsByAddress': grpc.unary_unary_rpc_method_handler(
          servicer.GetTransactionsByAddress,
          request_deserializer=qrl__pb2.GetTransactionsByAddressReq.FromString,
          response_serializer=qrl__pb2.GetTransactionsByAddressResp.SerializeToString,
      ),
      'GetTransaction': grpc.unary_unary_rpc_method_handler(
          servicer.GetTransaction,
          request_deserializer=qrl__pb2.GetTransactionReq.FromString,
          response_serializer=qrl__pb2.GetTransactionResp.SerializeToString,
      ),
      'GetBalance': grpc.unary_unary_rpc_method_handler(
          servicer.GetBalance,
          request_deserializer=qrl__pb2.GetBalanceReq.FromString,
          response_serializer=qrl__pb2.GetBalanceResp.SerializeToString,
      ),
      'GetOTS': grpc.unary_unary_rpc_method_handler(
          servicer.GetOTS,
          request_deserializer=qrl__pb2.GetOTSReq.FromString,
          response_serializer=qrl__pb2.GetOTSResp.SerializeToString,
      ),
      'GetHeight': grpc.unary_unary_rpc_method_handler(
          servicer.GetHeight,
          request_deserializer=qrl__pb2.GetHeightReq.FromString,
          response_serializer=qrl__pb2.GetHeightResp.SerializeToString,
      ),
      'GetBlock': grpc.unary_unary_rpc_method_handler(
          servicer.GetBlock,
          request_deserializer=qrl__pb2.GetBlockReq.FromString,
          response_serializer=qrl__pb2.GetBlockResp.SerializeToString,
      ),
      'GetBlockByNumber': grpc.unary_unary_rpc_method_handler(
          servicer.GetBlockByNumber,
          request_deserializer=qrl__pb2.GetBlockByNumberReq.FromString,
          response_serializer=qrl__pb2.GetBlockByNumberResp.SerializeToString,
      ),
  }
  generic_handler = grpc.method_handlers_generic_handler(
      'qrl.PublicAPI', rpc_method_handlers)
  server.add_generic_rpc_handlers((generic_handler,))


class AdminAPIStub(object):
  """This is a place holder for testing/instrumentation APIs
  """

  def __init__(self, channel):
    """Constructor.

    Args:
      channel: A grpc.Channel.
    """


class AdminAPIServicer(object):
  """This is a place holder for testing/instrumentation APIs
  """


def add_AdminAPIServicer_to_server(servicer, server):
  rpc_method_handlers = {
  }
  generic_handler = grpc.method_handlers_generic_handler(
      'qrl.AdminAPI', rpc_method_handlers)
  server.add_generic_rpc_handlers((generic_handler,))



# coding=utf-8
# Distributed under the MIT software license, see the accompanying
# file LICENSE or http://www.opensource.org/licenses/mit-license.php.
from qrl.core import config
from qrl.core.misc import ntp


class OutgoingMessage:
    def __init__(self, priority, message):
        self.priority = priority
        self.timestamp = int(ntp.getTime())
        self.message = message

    def is_expired(self):
        return self.timestamp - ntp.getTime() > config.user.outgoing_message_expiry

    def __lt__(self, outgoing_message):
        return self.priority < outgoing_message.priority



import os

from qrl.core.ChainManager import ChainManager
from qrl.core.GenesisBlock import GenesisBlock
from qrl.core.State import State
from pyqryptonight.pyqryptonight import PoWHelper, StringToUInt256, UInt256ToString


def main():
    persistent_state = State()
    chain_manager = ChainManager(state=persistent_state)
    chain_manager.load(GenesisBlock())

    ph = PoWHelper()
    difficulty = StringToUInt256('5000')

    filename = os.path.expanduser("~/crypto/qryptonight/modeling/blockdata.csv")

    with open(filename, 'w') as f:
        f.write("i,timestamp,prev_timestamp,delta,difficulty,target\n")
        prev_timestamp = None
        for i in range(chain_manager.height):
            block = chain_manager.get_block_by_number(i)

            if i == 0:
                prev_timestamp = block.blockheader.timestamp
                continue

            target = ph.getTarget(difficulty)
            delta = block.blockheader.timestamp - prev_timestamp

            outs = "{},{},{},{},{},{}\n".format(i,
                                                block.blockheader.timestamp,
                                                prev_timestamp,
                                                delta,
                                                UInt256ToString(difficulty),
                                                UInt256ToString(target))

            f.write(outs)

            difficulty = ph.getDifficulty(block.blockheader.timestamp, prev_timestamp, difficulty)
            difficulty = StringToUInt256(str(max(2, int(UInt256ToString(difficulty)))))
            prev_timestamp = block.blockheader.timestamp


if __name__ == '__main__':
    main()



# coding=utf-8
# Distributed under the MIT software license, see the accompanying
# file LICENSE or http://www.opensource.org/licenses/mit-license.php.
from unittest import TestCase

from qrl.core import config
from qrl.core.misc import logger
from qrl.core.GenesisBlock import GenesisBlock
from tests.misc.helper import clean_genesis

logger.initialize_default()


class TestGenesisBlock(TestCase):
    def __init__(self, *args, **kwargs):
        super(TestGenesisBlock, self).__init__(*args, **kwargs)

    def test_genesis_block_values(self):
        with clean_genesis():
            gb = GenesisBlock()

            self.assertIsNotNone(gb)
            self.assertEqual(0, gb.block_number)
            self.assertEqual(config.user.genesis_prev_headerhash, gb.prev_headerhash)
            self.assertEqual(1, len(gb.genesis_balance))



# coding=utf-8
# Distributed under the MIT software license, see the accompanying
# file LICENSE or http://www.opensource.org/licenses/mit-license.php.
from pyqrllib.pyqrllib import bin2hstr
from twisted.internet.task import cooperate

from qrl.core import ChainManager
from qrl.core.TransactionPool import TransactionPool
from qrl.core.misc import logger
from qrl.core.txs.Transaction import Transaction


class TxnProcessor:
    def __init__(self,
                 chain_manager: ChainManager,
                 transaction_pool_obj: TransactionPool,
                 broadcast_tx):
        self.chain_manager = chain_manager
        self.transaction_pool_obj = transaction_pool_obj
        self.broadcast_tx = broadcast_tx

    def __iter__(self):
        return self

    def __next__(self):
        tx_timestamp = self.transaction_pool_obj.get_pending_transaction()

        if not tx_timestamp:
            raise StopIteration

        tx, timestamp = tx_timestamp

        if not tx.validate():
            return False

        addr_from_state = self.chain_manager.get_address_state(address=tx.addr_from)
        addr_from_pk_state = addr_from_state

        addr_from_pk = Transaction.get_slave(tx)
        if addr_from_pk:
            addr_from_pk_state = self.chain_manager.get_address_state(address=addr_from_pk)

        is_valid_state = tx.validate_extended(addr_from_state=addr_from_state,
                                              addr_from_pk_state=addr_from_pk_state)

        if not is_valid_state:
            logger.info('>>>TX %s failed is_valid_state', bin2hstr(tx.txhash))
            return False

        is_valid_pool_state = tx.validate_transaction_pool(self.transaction_pool_obj.transaction_pool)

        if not is_valid_pool_state:
            logger.info('>>>TX %s failed is_valid_pool_state', bin2hstr(tx.txhash))
            return False

        logger.info('A TXN has been Processed %s', bin2hstr(tx.txhash))
        self.transaction_pool_obj.add_tx_to_pool(tx, self.chain_manager.last_block.block_number, timestamp)
        self.broadcast_tx(tx)

        return True

    @staticmethod
    def iterator(iterObj):
        for _ in iterObj:
            yield None

    @staticmethod
    def create_cooperate(iterObj):
        return cooperate(TxnProcessor.iterator(iterObj))



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
import grpc

import qrl.generated.qrlwallet_pb2 as qrlwallet__pb2


class WalletAPIStub(object):
  """//////////////////////////
  //////////////////////////
  //////////////////////////
  ////     API       ///////
  //////////////////////////
  //////////////////////////
  //////////////////////////

  This service describes the Wallet API
  """

  def __init__(self, channel):
    """Constructor.

    Args:
      channel: A grpc.Channel.
    """
    self.AddNewAddress = channel.unary_unary(
        '/qrl.WalletAPI/AddNewAddress',
        request_serializer=qrlwallet__pb2.AddNewAddressReq.SerializeToString,
        response_deserializer=qrlwallet__pb2.AddNewAddressResp.FromString,
        )
    self.AddNewAddressWithSlaves = channel.unary_unary(
        '/qrl.WalletAPI/AddNewAddressWithSlaves',
        request_serializer=qrlwallet__pb2.AddNewAddressWithSlavesReq.SerializeToString,
        response_deserializer=qrlwallet__pb2.AddNewAddressResp.FromString,
        )
    self.ListAddresses = channel.unary_unary(
        '/qrl.WalletAPI/ListAddresses',
        request_serializer=qrlwallet__pb2.ListAddressesReq.SerializeToString,
        response_deserializer=qrlwallet__pb2.ListAddressesResp.FromString,
        )
    self.RemoveAddress = channel.unary_unary(
        '/qrl.WalletAPI/RemoveAddress',
        request_serializer=qrlwallet__pb2.RemoveAddressReq.SerializeToString,
        response_deserializer=qrlwallet__pb2.RemoveAddressResp.FromString,
        )
    self.IsValidAddress = channel.unary_unary(
        '/qrl.WalletAPI/IsValidAddress',
        request_serializer=qrlwallet__pb2.ValidAddressReq.SerializeToString,
        response_deserializer=qrlwallet__pb2.ValidAddressResp.FromString,
        )
    self.EncryptWallet = channel.unary_unary(
        '/qrl.WalletAPI/EncryptWallet',
        request_serializer=qrlwallet__pb2.EncryptWalletReq.SerializeToString,
        response_deserializer=qrlwallet__pb2.EncryptWalletResp.FromString,
        )
    self.LockWallet = channel.unary_unary(
        '/qrl.WalletAPI/LockWallet',
        request_serializer=qrlwallet__pb2.LockWalletReq.SerializeToString,
        response_deserializer=qrlwallet__pb2.LockWalletResp.FromString,
        )
    self.UnlockWallet = channel.unary_unary(
        '/qrl.WalletAPI/UnlockWallet',
        request_serializer=qrlwallet__pb2.UnlockWalletReq.SerializeToString,
        response_deserializer=qrlwallet__pb2.UnlockWalletResp.FromString,
        )
    self.GetRecoverySeeds = channel.unary_unary(
        '/qrl.WalletAPI/GetRecoverySeeds',
        request_serializer=qrlwallet__pb2.GetRecoverySeedsReq.SerializeToString,
        response_deserializer=qrlwallet__pb2.GetRecoverySeedsResp.FromString,
        )
    self.GetWalletInfo = channel.unary_unary(
        '/qrl.WalletAPI/GetWalletInfo',
        request_serializer=qrlwallet__pb2.GetWalletInfoReq.SerializeToString,
        response_deserializer=qrlwallet__pb2.GetWalletInfoResp.FromString,
        )
    self.RelayTransferTxn = channel.unary_unary(
        '/qrl.WalletAPI/RelayTransferTxn',
        request_serializer=qrlwallet__pb2.RelayTransferTxnReq.SerializeToString,
        response_deserializer=qrlwallet__pb2.RelayTxnResp.FromString,
        )
    self.RelayTransferTxnBySlave = channel.unary_unary(
        '/qrl.WalletAPI/RelayTransferTxnBySlave',
        request_serializer=qrlwallet__pb2.RelayTransferTxnBySlaveReq.SerializeToString,
        response_deserializer=qrlwallet__pb2.RelayTxnResp.FromString,
        )
    self.RelayMessageTxn = channel.unary_unary(
        '/qrl.WalletAPI/RelayMessageTxn',
        request_serializer=qrlwallet__pb2.RelayMessageTxnReq.SerializeToString,
        response_deserializer=qrlwallet__pb2.RelayTxnResp.FromString,
        )
    self.RelayMessageTxnBySlave = channel.unary_unary(
        '/qrl.WalletAPI/RelayMessageTxnBySlave',
        request_serializer=qrlwallet__pb2.RelayMessageTxnBySlaveReq.SerializeToString,
        response_deserializer=qrlwallet__pb2.RelayTxnResp.FromString,
        )
    self.RelayTokenTxn = channel.unary_unary(
        '/qrl.WalletAPI/RelayTokenTxn',
        request_serializer=qrlwallet__pb2.RelayTokenTxnReq.SerializeToString,
        response_deserializer=qrlwallet__pb2.RelayTxnResp.FromString,
        )
    self.RelayTokenTxnBySlave = channel.unary_unary(
        '/qrl.WalletAPI/RelayTokenTxnBySlave',
        request_serializer=qrlwallet__pb2.RelayTokenTxnBySlaveReq.SerializeToString,
        response_deserializer=qrlwallet__pb2.RelayTxnResp.FromString,
        )
    self.RelayTransferTokenTxn = channel.unary_unary(
        '/qrl.WalletAPI/RelayTransferTokenTxn',
        request_serializer=qrlwallet__pb2.RelayTransferTokenTxnReq.SerializeToString,
        response_deserializer=qrlwallet__pb2.RelayTxnResp.FromString,
        )
    self.RelayTransferTokenTxnBySlave = channel.unary_unary(
        '/qrl.WalletAPI/RelayTransferTokenTxnBySlave',
        request_serializer=qrlwallet__pb2.RelayTransferTokenTxnBySlaveReq.SerializeToString,
        response_deserializer=qrlwallet__pb2.RelayTxnResp.FromString,
        )
    self.RelaySlaveTxn = channel.unary_unary(
        '/qrl.WalletAPI/RelaySlaveTxn',
        request_serializer=qrlwallet__pb2.RelaySlaveTxnReq.SerializeToString,
        response_deserializer=qrlwallet__pb2.RelayTxnResp.FromString,
        )
    self.RelaySlaveTxnBySlave = channel.unary_unary(
        '/qrl.WalletAPI/RelaySlaveTxnBySlave',
        request_serializer=qrlwallet__pb2.RelaySlaveTxnBySlaveReq.SerializeToString,
        response_deserializer=qrlwallet__pb2.RelayTxnResp.FromString,
        )
    self.ChangePassphrase = channel.unary_unary(
        '/qrl.WalletAPI/ChangePassphrase',
        request_serializer=qrlwallet__pb2.ChangePassphraseReq.SerializeToString,
        response_deserializer=qrlwallet__pb2.ChangePassphraseResp.FromString,
        )
    self.GetTransactionsByAddress = channel.unary_unary(
        '/qrl.WalletAPI/GetTransactionsByAddress',
        request_serializer=qrlwallet__pb2.TransactionsByAddressReq.SerializeToString,
        response_deserializer=qrlwallet__pb2.TransactionsByAddressResp.FromString,
        )
    self.GetTransaction = channel.unary_unary(
        '/qrl.WalletAPI/GetTransaction',
        request_serializer=qrlwallet__pb2.TransactionReq.SerializeToString,
        response_deserializer=qrlwallet__pb2.TransactionResp.FromString,
        )
    self.GetBalance = channel.unary_unary(
        '/qrl.WalletAPI/GetBalance',
        request_serializer=qrlwallet__pb2.BalanceReq.SerializeToString,
        response_deserializer=qrlwallet__pb2.BalanceResp.FromString,
        )
    self.GetOTS = channel.unary_unary(
        '/qrl.WalletAPI/GetOTS',
        request_serializer=qrlwallet__pb2.OTSReq.SerializeToString,
        response_deserializer=qrlwallet__pb2.OTSResp.FromString,
        )
    self.GetHeight = channel.unary_unary(
        '/qrl.WalletAPI/GetHeight',
        request_serializer=qrlwallet__pb2.HeightReq.SerializeToString,
        response_deserializer=qrlwallet__pb2.HeightResp.FromString,
        )
    self.GetBlock = channel.unary_unary(
        '/qrl.WalletAPI/GetBlock',
        request_serializer=qrlwallet__pb2.BlockReq.SerializeToString,
        response_deserializer=qrlwallet__pb2.BlockResp.FromString,
        )
    self.GetBlockByNumber = channel.unary_unary(
        '/qrl.WalletAPI/GetBlockByNumber',
        request_serializer=qrlwallet__pb2.BlockByNumberReq.SerializeToString,
        response_deserializer=qrlwallet__pb2.BlockResp.FromString,
        )
    self.GetAddressFromPK = channel.unary_unary(
        '/qrl.WalletAPI/GetAddressFromPK',
        request_serializer=qrlwallet__pb2.AddressFromPKReq.SerializeToString,
        response_deserializer=qrlwallet__pb2.AddressFromPKResp.FromString,
        )
    self.GetNodeInfo = channel.unary_unary(
        '/qrl.WalletAPI/GetNodeInfo',
        request_serializer=qrlwallet__pb2.NodeInfoReq.SerializeToString,
        response_deserializer=qrlwallet__pb2.NodeInfoResp.FromString,
        )


class WalletAPIServicer(object):
  """//////////////////////////
  //////////////////////////
  //////////////////////////
  ////     API       ///////
  //////////////////////////
  //////////////////////////
  //////////////////////////

  This service describes the Wallet API
  """

  def AddNewAddress(self, request, context):
    # missing associated documentation comment in .proto file
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def AddNewAddressWithSlaves(self, request, context):
    # missing associated documentation comment in .proto file
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def ListAddresses(self, request, context):
    """rpc AddAddressFromSeed(AddAddressFromSeedReq) returns (AddAddressFromSeedResp);

    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def RemoveAddress(self, request, context):
    # missing associated documentation comment in .proto file
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def IsValidAddress(self, request, context):
    # missing associated documentation comment in .proto file
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def EncryptWallet(self, request, context):
    # missing associated documentation comment in .proto file
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def LockWallet(self, request, context):
    # missing associated documentation comment in .proto file
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def UnlockWallet(self, request, context):
    # missing associated documentation comment in .proto file
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetRecoverySeeds(self, request, context):
    # missing associated documentation comment in .proto file
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetWalletInfo(self, request, context):
    # missing associated documentation comment in .proto file
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def RelayTransferTxn(self, request, context):
    # missing associated documentation comment in .proto file
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def RelayTransferTxnBySlave(self, request, context):
    # missing associated documentation comment in .proto file
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def RelayMessageTxn(self, request, context):
    # missing associated documentation comment in .proto file
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def RelayMessageTxnBySlave(self, request, context):
    # missing associated documentation comment in .proto file
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def RelayTokenTxn(self, request, context):
    # missing associated documentation comment in .proto file
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def RelayTokenTxnBySlave(self, request, context):
    # missing associated documentation comment in .proto file
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def RelayTransferTokenTxn(self, request, context):
    # missing associated documentation comment in .proto file
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def RelayTransferTokenTxnBySlave(self, request, context):
    # missing associated documentation comment in .proto file
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def RelaySlaveTxn(self, request, context):
    # missing associated documentation comment in .proto file
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def RelaySlaveTxnBySlave(self, request, context):
    # missing associated documentation comment in .proto file
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def ChangePassphrase(self, request, context):
    # missing associated documentation comment in .proto file
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetTransactionsByAddress(self, request, context):
    # missing associated documentation comment in .proto file
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetTransaction(self, request, context):
    # missing associated documentation comment in .proto file
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetBalance(self, request, context):
    # missing associated documentation comment in .proto file
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetOTS(self, request, context):
    # missing associated documentation comment in .proto file
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetHeight(self, request, context):
    # missing associated documentation comment in .proto file
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetBlock(self, request, context):
    # missing associated documentation comment in .proto file
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetBlockByNumber(self, request, context):
    # missing associated documentation comment in .proto file
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetAddressFromPK(self, request, context):
    # missing associated documentation comment in .proto file
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetNodeInfo(self, request, context):
    # missing associated documentation comment in .proto file
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')


def add_WalletAPIServicer_to_server(servicer, server):
  rpc_method_handlers = {
      'AddNewAddress': grpc.unary_unary_rpc_method_handler(
          servicer.AddNewAddress,
          request_deserializer=qrlwallet__pb2.AddNewAddressReq.FromString,
          response_serializer=qrlwallet__pb2.AddNewAddressResp.SerializeToString,
      ),
      'AddNewAddressWithSlaves': grpc.unary_unary_rpc_method_handler(
          servicer.AddNewAddressWithSlaves,
          request_deserializer=qrlwallet__pb2.AddNewAddressWithSlavesReq.FromString,
          response_serializer=qrlwallet__pb2.AddNewAddressResp.SerializeToString,
      ),
      'ListAddresses': grpc.unary_unary_rpc_method_handler(
          servicer.ListAddresses,
          request_deserializer=qrlwallet__pb2.ListAddressesReq.FromString,
          response_serializer=qrlwallet__pb2.ListAddressesResp.SerializeToString,
      ),
      'RemoveAddress': grpc.unary_unary_rpc_method_handler(
          servicer.RemoveAddress,
          request_deserializer=qrlwallet__pb2.RemoveAddressReq.FromString,
          response_serializer=qrlwallet__pb2.RemoveAddressResp.SerializeToString,
      ),
      'IsValidAddress': grpc.unary_unary_rpc_method_handler(
          servicer.IsValidAddress,
          request_deserializer=qrlwallet__pb2.ValidAddressReq.FromString,
          response_serializer=qrlwallet__pb2.ValidAddressResp.SerializeToString,
      ),
      'EncryptWallet': grpc.unary_unary_rpc_method_handler(
          servicer.EncryptWallet,
          request_deserializer=qrlwallet__pb2.EncryptWalletReq.FromString,
          response_serializer=qrlwallet__pb2.EncryptWalletResp.SerializeToString,
      ),
      'LockWallet': grpc.unary_unary_rpc_method_handler(
          servicer.LockWallet,
          request_deserializer=qrlwallet__pb2.LockWalletReq.FromString,
          response_serializer=qrlwallet__pb2.LockWalletResp.SerializeToString,
      ),
      'UnlockWallet': grpc.unary_unary_rpc_method_handler(
          servicer.UnlockWallet,
          request_deserializer=qrlwallet__pb2.UnlockWalletReq.FromString,
          response_serializer=qrlwallet__pb2.UnlockWalletResp.SerializeToString,
      ),
      'GetRecoverySeeds': grpc.unary_unary_rpc_method_handler(
          servicer.GetRecoverySeeds,
          request_deserializer=qrlwallet__pb2.GetRecoverySeedsReq.FromString,
          response_serializer=qrlwallet__pb2.GetRecoverySeedsResp.SerializeToString,
      ),
      'GetWalletInfo': grpc.unary_unary_rpc_method_handler(
          servicer.GetWalletInfo,
          request_deserializer=qrlwallet__pb2.GetWalletInfoReq.FromString,
          response_serializer=qrlwallet__pb2.GetWalletInfoResp.SerializeToString,
      ),
      'RelayTransferTxn': grpc.unary_unary_rpc_method_handler(
          servicer.RelayTransferTxn,
          request_deserializer=qrlwallet__pb2.RelayTransferTxnReq.FromString,
          response_serializer=qrlwallet__pb2.RelayTxnResp.SerializeToString,
      ),
      'RelayTransferTxnBySlave': grpc.unary_unary_rpc_method_handler(
          servicer.RelayTransferTxnBySlave,
          request_deserializer=qrlwallet__pb2.RelayTransferTxnBySlaveReq.FromString,
          response_serializer=qrlwallet__pb2.RelayTxnResp.SerializeToString,
      ),
      'RelayMessageTxn': grpc.unary_unary_rpc_method_handler(
          servicer.RelayMessageTxn,
          request_deserializer=qrlwallet__pb2.RelayMessageTxnReq.FromString,
          response_serializer=qrlwallet__pb2.RelayTxnResp.SerializeToString,
      ),
      'RelayMessageTxnBySlave': grpc.unary_unary_rpc_method_handler(
          servicer.RelayMessageTxnBySlave,
          request_deserializer=qrlwallet__pb2.RelayMessageTxnBySlaveReq.FromString,
          response_serializer=qrlwallet__pb2.RelayTxnResp.SerializeToString,
      ),
      'RelayTokenTxn': grpc.unary_unary_rpc_method_handler(
          servicer.RelayTokenTxn,
          request_deserializer=qrlwallet__pb2.RelayTokenTxnReq.FromString,
          response_serializer=qrlwallet__pb2.RelayTxnResp.SerializeToString,
      ),
      'RelayTokenTxnBySlave': grpc.unary_unary_rpc_method_handler(
          servicer.RelayTokenTxnBySlave,
          request_deserializer=qrlwallet__pb2.RelayTokenTxnBySlaveReq.FromString,
          response_serializer=qrlwallet__pb2.RelayTxnResp.SerializeToString,
      ),
      'RelayTransferTokenTxn': grpc.unary_unary_rpc_method_handler(
          servicer.RelayTransferTokenTxn,
          request_deserializer=qrlwallet__pb2.RelayTransferTokenTxnReq.FromString,
          response_serializer=qrlwallet__pb2.RelayTxnResp.SerializeToString,
      ),
      'RelayTransferTokenTxnBySlave': grpc.unary_unary_rpc_method_handler(
          servicer.RelayTransferTokenTxnBySlave,
          request_deserializer=qrlwallet__pb2.RelayTransferTokenTxnBySlaveReq.FromString,
          response_serializer=qrlwallet__pb2.RelayTxnResp.SerializeToString,
      ),
      'RelaySlaveTxn': grpc.unary_unary_rpc_method_handler(
          servicer.RelaySlaveTxn,
          request_deserializer=qrlwallet__pb2.RelaySlaveTxnReq.FromString,
          response_serializer=qrlwallet__pb2.RelayTxnResp.SerializeToString,
      ),
      'RelaySlaveTxnBySlave': grpc.unary_unary_rpc_method_handler(
          servicer.RelaySlaveTxnBySlave,
          request_deserializer=qrlwallet__pb2.RelaySlaveTxnBySlaveReq.FromString,
          response_serializer=qrlwallet__pb2.RelayTxnResp.SerializeToString,
      ),
      'ChangePassphrase': grpc.unary_unary_rpc_method_handler(
          servicer.ChangePassphrase,
          request_deserializer=qrlwallet__pb2.ChangePassphraseReq.FromString,
          response_serializer=qrlwallet__pb2.ChangePassphraseResp.SerializeToString,
      ),
      'GetTransactionsByAddress': grpc.unary_unary_rpc_method_handler(
          servicer.GetTransactionsByAddress,
          request_deserializer=qrlwallet__pb2.TransactionsByAddressReq.FromString,
          response_serializer=qrlwallet__pb2.TransactionsByAddressResp.SerializeToString,
      ),
      'GetTransaction': grpc.unary_unary_rpc_method_handler(
          servicer.GetTransaction,
          request_deserializer=qrlwallet__pb2.TransactionReq.FromString,
          response_serializer=qrlwallet__pb2.TransactionResp.SerializeToString,
      ),
      'GetBalance': grpc.unary_unary_rpc_method_handler(
          servicer.GetBalance,
          request_deserializer=qrlwallet__pb2.BalanceReq.FromString,
          response_serializer=qrlwallet__pb2.BalanceResp.SerializeToString,
      ),
      'GetOTS': grpc.unary_unary_rpc_method_handler(
          servicer.GetOTS,
          request_deserializer=qrlwallet__pb2.OTSReq.FromString,
          response_serializer=qrlwallet__pb2.OTSResp.SerializeToString,
      ),
      'GetHeight': grpc.unary_unary_rpc_method_handler(
          servicer.GetHeight,
          request_deserializer=qrlwallet__pb2.HeightReq.FromString,
          response_serializer=qrlwallet__pb2.HeightResp.SerializeToString,
      ),
      'GetBlock': grpc.unary_unary_rpc_method_handler(
          servicer.GetBlock,
          request_deserializer=qrlwallet__pb2.BlockReq.FromString,
          response_serializer=qrlwallet__pb2.BlockResp.SerializeToString,
      ),
      'GetBlockByNumber': grpc.unary_unary_rpc_method_handler(
          servicer.GetBlockByNumber,
          request_deserializer=qrlwallet__pb2.BlockByNumberReq.FromString,
          response_serializer=qrlwallet__pb2.BlockResp.SerializeToString,
      ),
      'GetAddressFromPK': grpc.unary_unary_rpc_method_handler(
          servicer.GetAddressFromPK,
          request_deserializer=qrlwallet__pb2.AddressFromPKReq.FromString,
          response_serializer=qrlwallet__pb2.AddressFromPKResp.SerializeToString,
      ),
      'GetNodeInfo': grpc.unary_unary_rpc_method_handler(
          servicer.GetNodeInfo,
          request_deserializer=qrlwallet__pb2.NodeInfoReq.FromString,
          response_serializer=qrlwallet__pb2.NodeInfoResp.SerializeToString,
      ),
  }
  generic_handler = grpc.method_handlers_generic_handler(
      'qrl.WalletAPI', rpc_method_handlers)
  server.add_generic_rpc_handlers((generic_handler,))



# coding=utf-8
# Distributed under the MIT software license, see the accompanying
# file LICENSE or http://www.opensource.org/licenses/mit-license.php.
import json
import os
import shutil
import tempfile
from unittest import TestCase, mock

from click.testing import CliRunner

from qrl.cli import qrl as qrl_cli
from qrl.core import config
from qrl.core.misc import logger
from qrl.generated import qrl_pb2

logger.initialize_default()

unsigned_tx = b"\nOQ010200be640405bb61d104e329cc94b6807b6e713ef0dea80aa5aa73abff8dd88946b41172569c\x1aC\x01\x02\x00\x80\x9dg/U\xb1N\xf2_\x0e~j%\xb2\x15\x05\xa7y\x19\x8f\xc0>\x05`\x90\xe3>\xaa\x9a(\xd3\xc7U\x91\xbab\x90{\xaa^\xadQ\xca\xbf\xd3\xbc\xd9\x93\xf0:D\xca\xd8v\x97\x08\xa8x\x9c-\n4\xd6e:,\n'\x01\x04\x00\xd9\xf1\xef\xe5\xb2r\xe0B\xdc\xc8\xefi\x0f\x0e\x90\xca\x8b\x0bn\xdb\xa0\xd2o\x81\xe7\xaf\xf1*gT\xb2\x17\x88\x16\x9f\x7f\x12\x01\x00"  # noqa
bin2hstr_unsigned_tx = '1a4301030038ea6375069f8272cc1a6601b3c76c21519455603d370036b97c779ada3565854e3983bd564298c49ae2e7fa6e28d4b954d8cd59398f1225b08d6144854aee0e3a8c010a2701050000a31343e48abce29464f063b30827ecf2c552743cab13d4f3b457b55410b548a0f8bca60a270105009ca50ed86497e6b2cfcca8c525191c741220eacf79a697e078bf9b53a9899b17b3f839d10a27010500e135decb3328a27e51c47064df2dbbe799e79812291cc5e7cfad08a82a62d64e4fa813aa120f8094ebdc0380a8d6b90780bcc1960b'  # noqa
bin2hstr_unsigned_tx2 = '0a2701050000a31343e48abce29464f063b30827ecf2c552743cab13d4f3b457b55410b548a0f8bca61a4301030038ea6375069f8272cc1a6601b3c76c21519455603d370036b97c779ada3565854e3983bd564298c49ae2e7fa6e28d4b954d8cd59398f1225b08d6144854aee0e3a2c0a270105009ca50ed86497e6b2cfcca8c525191c741220eacf79a697e078bf9b53a9899b17b3f839d1120100'  # noqa
bin2hstr_signed_tx = ' '  # noqa

qaddr_1 = "Q01050000a31343e48abce29464f063b30827ecf2c552743cab13d4f3b457b55410b548a0f8bca6"
qaddr_2 = "Q0105009ca50ed86497e6b2cfcca8c525191c741220eacf79a697e078bf9b53a9899b17b3f839d1"
qaddr_3 = "Q010500e135decb3328a27e51c47064df2dbbe799e79812291cc5e7cfad08a82a62d64e4fa813aa"
qaddr_4 = "Q0105001fa3a2038c0b4b947cccbf589d53fc7749f0a04f118ec1ffadedd1b4db37e83c513aa259"


def open_wallet(filename="wallet.json"):
    with open(filename) as f:
        return json.load(f)


@mock.patch('qrl.cli.grpc.insecure_channel', new=mock.MagicMock())
class TestCLI_Wallet_Gen(TestCase):
    def __init__(self, *args, **kwargs):
        super(TestCLI_Wallet_Gen, self).__init__(*args, **kwargs)

    def setUp(self):
        self.runner = CliRunner(env={"LC_ALL": "C.UTF-8", "LANG": "C.UTF-8"})
        self.prev_dir = os.getcwd()
        self.temp_dir = tempfile.mkdtemp()

        os.chdir(self.temp_dir)

    def tearDown(self):
        del self.runner  # running this test suite often results in leaked pipe handles. This could mitigate that.
        os.chdir(self.prev_dir)
        shutil.rmtree(self.temp_dir)

    def wallet_gen_default_height(self):
        result = self.runner.invoke(qrl_cli, ["wallet_gen"])
        wallet = open_wallet()
        self.assertIn(self.temp_dir, result.output)
        self.assertEqual(wallet["addresses"][0]["height"], 12)

    def test_wallet_gen_different_height(self):
        result = self.runner.invoke(qrl_cli, ["wallet_gen", "--height=4"])
        wallet = open_wallet()
        self.assertIn(self.temp_dir, result.output)
        self.assertEqual(wallet["addresses"][0]["height"], 4)

    def test_wallet_gen_different_hash_function(self):
        result = self.runner.invoke(qrl_cli, ["wallet_gen", "--height=4", "--hash_function=sha2_256"])
        wallet = open_wallet()
        self.assertIn(self.temp_dir, result.output)
        self.assertEqual(wallet["addresses"][0]["hashFunction"], "sha2_256")

    def test_wallet_gen_json(self):
        result = self.runner.invoke(qrl_cli, ["--json", "wallet_gen", "--height=4"])
        self.assertTrue(json.loads(result.output))  # Throws an exception if output is not valid JSON

    def test_wallet_gen_encrypt(self):
        result = self.runner.invoke(qrl_cli, ["wallet_gen", "--height=4", "--encrypt"], input='password\npassword\n')
        wallet = open_wallet()
        self.assertIn(self.temp_dir, result.output)
        self.assertTrue(wallet["encrypted"])


@mock.patch('qrl.cli.grpc.insecure_channel', new=mock.MagicMock())
class TestCLI(TestCase):
    def __init__(self, *args, **kwargs):
        super(TestCLI, self).__init__(*args, **kwargs)

    def setUp(self):
        self.runner = CliRunner(env={"LC_ALL": "C.UTF-8", "LANG": "C.UTF-8"})
        self.prev_dir = os.getcwd()
        self.temp_dir = tempfile.mkdtemp()
        os.chdir(self.temp_dir)
        self.runner.invoke(qrl_cli, ["wallet_gen", "--height=4"])

    def tearDown(self):
        os.chdir(self.prev_dir)
        shutil.rmtree(self.temp_dir)

    def test_wallet_ls(self):
        result = self.runner.invoke(qrl_cli, ["wallet_ls"])
        wallet = open_wallet()
        self.assertIn(wallet["addresses"][0]["address"], result.output)
        self.assertIn(self.temp_dir, result.output)  # You should know which wallet you've opened.

    def test_wallet_ls_verbose(self):
        result = self.runner.invoke(qrl_cli, ["-v", "wallet_ls"])
        wallet = open_wallet()
        self.assertIn(wallet["addresses"][0]["hashFunction"], result.output)

    def test_wallet_ls_empty(self):
        os.remove("wallet.json")
        result = self.runner.invoke(qrl_cli, ["wallet_ls"])
        self.assertIn("No wallet found", result.output)

    def test_wallet_ls_json(self):
        result = self.runner.invoke(qrl_cli, ["--json", "wallet_ls"])
        wallet = open_wallet()
        self.assertTrue(json.loads(result.output))  # Throws an exception if output is not valid JSON
        self.assertIn(wallet["addresses"][0]["address"], result.output)
        self.assertIn(self.temp_dir, result.output)  # You should know which wallet you've opened.

    def test_wallet_add(self):
        result = self.runner.invoke(qrl_cli, ["wallet_add", "--height=4"])
        wallet = open_wallet()
        self.assertIn(wallet["addresses"][1]["address"], result.output)
        self.assertEqual(wallet["addresses"][1]["height"], 4)

    def test_wallet_add_inherit_encryption_status(self):
        self.runner.invoke(qrl_cli, ["wallet_encrypt"], input='password\npassword\n')
        self.runner.invoke(qrl_cli, ["wallet_add", "--height=4"], input='password\n')
        wallet = open_wallet()
        self.assertTrue(wallet["encrypted"])

    def test_wallet_add_different_hash_function(self):
        self.runner.invoke(qrl_cli, ["wallet_add", "--height=4", "--hash_function=shake256"])
        wallet = open_wallet()
        self.assertEqual(wallet["addresses"][0]["hashFunction"], "shake128")
        self.assertEqual(wallet["addresses"][1]["hashFunction"], "shake256")

    def test_wallet_recover_hexseed(self):
        os.rename("wallet.json", "wallet_orig.json")
        wallet_orig = open_wallet("wallet_orig.json")
        self.runner.invoke(qrl_cli, ["wallet_recover", "--seed-type=hexseed"],
                           input='\n'.join([wallet_orig["addresses"][0]["hexseed"], 'y']))
        wallet_recovered = open_wallet()
        self.assertEqual(wallet_recovered, wallet_orig)

    def test_wallet_recover_hexseed_invalid_input(self):
        os.rename("wallet.json", "wallet_orig.json")
        wallet_orig = open_wallet("wallet_orig.json")
        result = self.runner.invoke(qrl_cli, ["wallet_recover", "--seed-type=hexseed"],
                                    input='\n'.join([wallet_orig["addresses"][0]["hexseed"] + "deadbeef", 'y']))
        self.assertIn('Hexseed must be of only', result.output)
        self.assertFalse(os.path.exists("wallet.json"))

        # If the recovered address is already in the wallet, it should react in this way.
        self.runner.invoke(qrl_cli, ["wallet_recover", "--seed-type=hexseed"],
                           input='\n'.join([wallet_orig["addresses"][0]["hexseed"], 'y']))
        result = self.runner.invoke(qrl_cli, ["wallet_recover", "--seed-type=hexseed"],
                                    input='\n'.join([wallet_orig["addresses"][0]["hexseed"], 'y']))
        self.assertEqual(result.exit_code, 0)
        self.assertIn('Wallet Address is already in the wallet list', result.output.strip())

    def test_wallet_recover_mnemonic(self):
        os.rename("wallet.json", "wallet_orig.json")
        wallet_orig = open_wallet("wallet_orig.json")
        self.runner.invoke(qrl_cli, ["wallet_recover", "--seed-type=mnemonic"],
                           input='\n'.join([wallet_orig["addresses"][0]["mnemonic"], 'y']))
        wallet_recovered = open_wallet()
        self.assertEqual(wallet_recovered, wallet_orig)

    def test_wallet_recover_mnemonic_invalid_input(self):
        os.rename("wallet.json", "wallet_orig.json")
        wallet_orig = open_wallet("wallet_orig.json")
        result = self.runner.invoke(qrl_cli, ["wallet_recover", "--seed-type=mnemonic"],
                                    input='\n'.join([wallet_orig["addresses"][0]["mnemonic"] + " bad", 'y']))
        self.assertIn('Mnemonic seed must contain only 34 words', result.output)
        self.assertFalse(os.path.exists("wallet.json"))

        # If the recovered address is already in the wallet, it should react in this way.
        self.runner.invoke(qrl_cli, ["wallet_recover", "--seed-type=mnemonic"],
                           input='\n'.join([wallet_orig["addresses"][0]["mnemonic"], 'y']))
        result = self.runner.invoke(qrl_cli, ["wallet_recover", "--seed-type=mnemonic"],
                                    input='\n'.join([wallet_orig["addresses"][0]["mnemonic"], 'y']))
        self.assertEqual(result.exit_code, 0)
        self.assertIn('Wallet Address is already in the wallet list', result.output.strip())

    def test_wallet_secret(self):
        wallet = open_wallet()
        result = self.runner.invoke(qrl_cli, ["wallet_secret", "--wallet-idx=0"])
        self.assertIn(wallet["addresses"][0]["address"], result.output)
        self.assertIn(wallet["addresses"][0]["mnemonic"], result.output)
        self.assertIn(wallet["addresses"][0]["hexseed"], result.output)

    def test_wallet_secret_encrypted(self):
        wallet = open_wallet()
        self.runner.invoke(qrl_cli, ["wallet_encrypt"], input='password\npassword\n')
        result = self.runner.invoke(qrl_cli, ["wallet_secret", "--wallet-idx=0"], input='password\npassword\n')
        self.assertIn(wallet["addresses"][0]["address"], result.output)
        self.assertIn(wallet["addresses"][0]["mnemonic"], result.output)
        self.assertIn(wallet["addresses"][0]["hexseed"], result.output)

    def test_wallet_secret_encrypt_decrypt_wrong(self):
        wallet = open_wallet()
        self.runner.invoke(qrl_cli, ["wallet_encrypt"], input='password\npassword\n')
        result = self.runner.invoke(qrl_cli, ["wallet_secret", "--wallet-idx=0"], input='password\npassword\n')
        self.assertIn(wallet["addresses"][0]["address"], result.output)
        self.assertIn(wallet["addresses"][0]["mnemonic"], result.output)
        self.assertIn(wallet["addresses"][0]["hexseed"], result.output)

        result = self.runner.invoke(qrl_cli, ["wallet_decrypt"], input='wrong_password\nwrong_password\n')
        self.assertEquals(1, result.exit_code)
        # TODO: Add appropriate error matching

    def test_wallet_secret_invalid_input(self):
        result = self.runner.invoke(qrl_cli, ["wallet_secret", "--wallet-idx=2"])
        self.assertEqual(result.output.strip(), 'Wallet index not found 2')

    def test_wallet_rm(self):
        self.runner.invoke(qrl_cli, ["wallet_add", "--height=4"])
        wallet = open_wallet()
        result = self.runner.invoke(qrl_cli, ["wallet_rm", "--wallet-idx=1", "--skip-confirmation"], input='y\n')
        self.assertNotIn(wallet["addresses"][1]["address"], result.output)
        result = self.runner.invoke(qrl_cli, ["wallet_rm", "--wallet-idx=0"], input='y\n')
        self.assertIn("No wallet found", result.output)

    def test_wallet_rm_invalid_input(self):
        result = self.runner.invoke(qrl_cli, ["wallet_rm", "--wallet-idx=2", "--skip-confirmation"], input='y\n')
        self.assertEqual(result.output.strip(), 'Wallet index not found 2')

    @mock.patch('qrl.cli.qrl_pb2_grpc.PublicAPIStub', autospec=True)
    def test_state(self, mock_stub):
        m_node_state_resp_serialized_to_string = b'\n/\n\x160.61.3+ngbd52b1a.dirty\x10\x03\x18\x02 D(\xa7\x010\x83!B\tExcession'
        m_node_state_resp = qrl_pb2.GetNodeStateResp()
        m_node_state_resp.ParseFromString(m_node_state_resp_serialized_to_string)

        mock_stub_instance = mock.MagicMock(name='this should be qrl_pb2_grpc.PublicAPIStub(channel)')
        mock_stub_instance.GetNodeState.return_value = m_node_state_resp

        mock_stub.name = 'this should be qrl_pb2_grpc.PublicAPIStub'
        mock_stub.return_value = mock_stub_instance

        result = self.runner.invoke(qrl_cli, ["state"])
        self.assertEqual(result.exit_code, 0)
        self.assertIn('Excession', result.output)

    @mock.patch('qrl.cli.qrl_pb2_grpc.PublicAPIStub', autospec=True)
    def test_state_json(self, mock_stub):
        m_node_state_resp_serialized = b'\n/\n\x160.61.3+ngbd52b1a.dirty\x10\x03\x18\x02 D(\xa7\x010\x83!B\tExcession'
        m_node_state_resp = qrl_pb2.GetNodeStateResp()
        m_node_state_resp.ParseFromString(m_node_state_resp_serialized)

        mock_stub_instance = mock.MagicMock(name='this should be qrl_pb2_grpc.PublicAPIStub(channel)')
        mock_stub_instance.GetNodeState.return_value = m_node_state_resp

        mock_stub.name = 'this should be qrl_pb2_grpc.PublicAPIStub'
        mock_stub.return_value = mock_stub_instance

        result = self.runner.invoke(qrl_cli, ["--json", "state"])
        print(result.output)
        print(result.exc_info)
        self.assertEqual(result.exit_code, 0)

        the_output = json.loads(result.output)
        self.assertEqual(the_output["info"]["networkId"], "Excession")

    @mock.patch('qrl.cli.config', autospec=True)
    @mock.patch('qrl.cli.Transaction', autospec=True)
    @mock.patch('qrl.cli.qrl_pb2_grpc.PublicAPIStub', autospec=True)
    def test_slave_tx_generate(self, mock_stub, mock_transaction, mock_config):
        # Mock out xmss_tree_height so this test runs faster!
        # But first, we must copy all attributes from the real config into the mock.
        # autospec does not do this for us, only the specification.
        mock_config.configure_mock(**config.__dict__)
        mock_config.dev.xmss_tree_height = 4

        mock_tx_attrs = {"name": "a mock tx object", "to_json.return_value": "{}"}
        mock_tx = mock.MagicMock(**mock_tx_attrs)

        mock_transaction.name = 'mock Transaction class'
        mock_transaction.from_pbdata.return_value = mock_tx

        mock_slave_txn_resp = mock.MagicMock(name='mock slaveTxnResp')
        mock_slave_txn_resp.extended_transaction_unsigned.tx = bin2hstr_unsigned_tx

        mock_stub_instance = mock.MagicMock(name='this should be qrl_pb2_grpc.PublicAPIStub(channel)')
        mock_stub_instance.GetSlaveTxn.return_value = mock_slave_txn_resp

        mock_stub.name = 'this should be qrl_pb2_grpc.PublicAPIStub'
        mock_stub.return_value = mock_stub_instance

        wallet = open_wallet()
        master_address = wallet["addresses"][0]["address"]

        # Simplest use case
        result = self.runner.invoke(qrl_cli, [
            "slave_tx_generate", "--src=0", "--master={}".format(master_address),
            "--number_of_slaves=1", "--access_type=0", "--fee=0"
        ])
        print(result.output)
        print(result.exc_info)
        self.assertEqual(result.exit_code, 0)
        self.assertTrue(os.path.exists('slaves.json'))

        # Does it work with 5 slaves?
        os.remove('slaves.json')
        result = self.runner.invoke(qrl_cli, [
            "slave_tx_generate", "--src=0", "--master={}".format(master_address),
            "--number_of_slaves=5", "--access_type=0", "--fee=0"
        ])
        self.assertEqual(result.exit_code, 0)
        self.assertTrue(os.path.exists('slaves.json'))

        # access_type could be 0 or 1. test 1.
        os.remove('slaves.json')
        result = self.runner.invoke(qrl_cli, [
            "slave_tx_generate", "--src=0", "--master={}".format(master_address),
            "--number_of_slaves=5", "--access_type=1", "--fee=0"
        ])
        self.assertEqual(result.exit_code, 0)
        self.assertTrue(os.path.exists('slaves.json'))

    @mock.patch('qrl.cli.config', autospec=True)
    @mock.patch('qrl.cli.Transaction', autospec=True)
    @mock.patch('qrl.cli.qrl_pb2_grpc.PublicAPIStub', autospec=True)
    def test_slave_tx_generate_invalid_input(self, mock_stub, mock_Transaction, mock_config):
        mock_config.configure_mock(**config.__dict__)
        mock_config.dev.xmss_tree_height = 4

        mock_tx_attrs = {"name": "a mock tx object", "to_json.return_value": "{}"}
        mock_tx = mock.MagicMock(**mock_tx_attrs)

        mock_Transaction.name = 'mock Transaction class'
        mock_Transaction.from_pbdata.return_value = mock_tx

        mock_slaveTxnResp = mock.MagicMock(name='mock slaveTxnResp')
        mock_slaveTxnResp.extended_transaction_unsigned.tx = bin2hstr_unsigned_tx

        mock_stub_instance = mock.MagicMock(name='this should be qrl_pb2_grpc.PublicAPIStub(channel)')
        mock_stub_instance.GetSlaveTxn.return_value = mock_slaveTxnResp

        mock_stub.name = 'this should be qrl_pb2_grpc.PublicAPIStub'
        mock_stub.return_value = mock_stub_instance

        wallet = open_wallet()
        master_address = wallet["addresses"][0]["address"]

        # It shouldn't allow > 100 slaves.
        result = self.runner.invoke(qrl_cli, [
            "slave_tx_generate", "--src=0", "--master={}".format(master_address),
            "--number_of_slaves=101", "--access_type=0", "--fee=0"
        ])
        self.assertNotEqual(result.exit_code, 0)
        self.assertFalse(os.path.exists('slaves.json'))

        # # Access types other than 0, 1 shouldn't work
        # result = self.runner.invoke(qrl_cli, [
        #     "slave_tx_generate", "--src=0", "--master={}".format(master_address),
        #     "--number_of_slaves=1", "--access_type=2", "--fee=0"
        # ])
        # self.assertNotEqual(result.exit_code, 0)
        # self.assertFalse(os.path.exists('slaves.json'))

        # Negative fees shoudn't work.
        result = self.runner.invoke(qrl_cli, [
            "slave_tx_generate", "--src=0", "--master={}".format(master_address),
            "--number_of_slaves=1", "--access_type=0", "--fee=-1"
        ])
        self.assertNotEqual(result.exit_code, 0)
        self.assertFalse(os.path.exists('slaves.json'))

    def test_tx_inspect(self):
        result = self.runner.invoke(qrl_cli, ["tx_inspect", "--txblob", bin2hstr_unsigned_tx])
        self.assertTrue(json.loads(result.output))

    def test_tx_inspect_invalid_input(self):
        result = self.runner.invoke(qrl_cli, ["tx_inspect", "--txblob", bin2hstr_unsigned_tx[2:]])
        self.assertNotEqual(result.exit_code, 0)
        self.assertIn('is not valid', result.output)

    @mock.patch('qrl.cli.qrl_pb2_grpc.PublicAPIStub', autospec=True)
    def test_tx_push(self, mock_stub):
        mock_error_code = 'Error? What error? This is a test'
        mock_push_transaction_resp = mock.MagicMock(name='this should be pushTransactionResp',
                                                    error_code=mock_error_code)

        attrs = {"name": "this should be stub.PushTransaction",
                 "PushTransaction.return_value": mock_push_transaction_resp}
        mock_stub_instance = mock.MagicMock(**attrs)

        mock_stub.name = 'this should be qrl_pb2_grpc.PublicAPIStub'
        mock_stub.return_value = mock_stub_instance

        result = self.runner.invoke(qrl_cli, ["tx_push", "--txblob", bin2hstr_signed_tx])
        self.assertIn(mock_error_code, result.output)

    @mock.patch('qrl.cli.qrl_pb2_grpc.PublicAPIStub', autospec=True)
    def test_tx_push_invalid_input(self, mock_stub):
        mock_error_code = 'Error? What error? This is a test'
        mock_push_transaction_resp = mock.MagicMock(name='this should be pushTransactionResp',
                                                    error_code=mock_error_code)

        attrs = {"name": "this should be stub.PushTransaction",
                 "PushTransaction.return_value": mock_push_transaction_resp}
        mock_stub_instance = mock.MagicMock(**attrs)

        mock_stub.name = 'this should be qrl_pb2_grpc.PublicAPIStub'
        mock_stub.return_value = mock_stub_instance

        # txblob should already be signed
        result = self.runner.invoke(qrl_cli, ["tx_push", "--txblob", bin2hstr_unsigned_tx])
        self.assertEqual(result.exit_code, 1)
        self.assertIn('Signature missing', result.output)

        # It should choke on invalid txblobs
        result = self.runner.invoke(qrl_cli, ["tx_push", "--txblob", bin2hstr_signed_tx[2:]])
        self.assertEqual(result.exit_code, 1)
        self.assertEqual(result.output.strip(), 'tx blob is not valid')

    @mock.patch('qrl.cli.qrl_pb2_grpc.PublicAPIStub', autospec=True)
    def test_tx_transfer(self, mock_stub):
        tx_pbdata_serialized_to_string = b"\n\x02\\n\x1aC\x01\x02\x00\x80\x9dg/U\xb1N\xf2_\x0e~j%\xb2\x15\x05\xa7y\x19\x8f\xc0>\x05`\x90\xe3>\xaa\x9a(\xd3\xc7U\x91\xbab\x90{\xaa^\xadQ\xca\xbf\xd3\xbc\xd9\x93\xf0:D\xca\xd8v\x97\x08\xa8x\x9c-\n4\xd6e:,\n'\x01\x06\x00\x95O\x16\xafx\xe3\x94Y\rc\x7f\x10D\xcd\x9f\xaf<\xc7\xf4\xa2\x93f\xaa\r\x8c\xa3 \xe40\x0bZ\xfe\xb0xy\xbb\x12\x01\x00"  # noqa
        mock_tx = qrl_pb2.Transaction()
        mock_tx.ParseFromString(tx_pbdata_serialized_to_string)

        m_transfer_coins_resp = mock.MagicMock(name='a fake transferCoinsResp')
        m_transfer_coins_resp.extended_transaction_unsigned.tx = mock_tx
        m_push_transaction_resp = mock.MagicMock(name='a fake pushTransactionResp', error_code=3)

        attrs = {
            "name": "my fake stub",
            "TransferCoins.return_value": m_transfer_coins_resp,
            "PushTransaction.return_value": m_push_transaction_resp
        }
        mock_stub_instance = mock.MagicMock(**attrs)

        mock_stub.name = 'a fake qrl_pb2_grpc.PublicAPIStub'
        mock_stub.return_value = mock_stub_instance

        wallet = open_wallet()

        # Simplest use case
        result = self.runner.invoke(qrl_cli, [
            "tx_transfer",
            "--src=0",
            "--master=",
            "--dsts={}".format(qaddr_1),
            "--amounts=1",
            "--fee=0",
            "--ots_key_index=0"
        ])
        self.assertEqual(result.exit_code, 0)
        print(result.output)
        self.assertIn('a fake pushTransactionResp', result.output.strip())

        # Should work with src=Qaddress as well
        result = self.runner.invoke(qrl_cli, [
            "tx_transfer",
            "--src={}".format(wallet["addresses"][0]["address"]),
            "--master=",
            "--dsts={}".format(qaddr_1),
            "--amounts=1",
            "--fee=0",
            "--ots_key_index=0"
        ])
        self.assertEqual(result.exit_code, 0)
        self.assertIn('a fake pushTransactionResp', result.output.strip())

        # Master should also work with a Qaddress.
        result = self.runner.invoke(qrl_cli, [
            "tx_transfer",
            "--src={}".format(wallet["addresses"][0]["address"]),
            "--master={}".format(qaddr_2),
            "--dsts={}".format(qaddr_1),
            "--amounts=1",
            "--fee=0",
            "--ots_key_index=0"
        ])
        self.assertEqual(result.exit_code, 0)
        self.assertIn('a fake pushTransactionResp', result.output.strip())

        # Multiple dsts should work too
        dsts = [qaddr_1, qaddr_2, qaddr_3]
        amounts = ["1", "2", "3"]
        result = self.runner.invoke(qrl_cli, [
            "tx_transfer",
            "--src=0",
            "--master=",
            "--dsts={}".format(" ".join(dsts)),
            "--amounts={}".format(" ".join(amounts)),
            "--fee=0",
            "--ots_key_index=0"
        ])
        self.assertEqual(result.exit_code, 0)
        self.assertIn('a fake pushTransactionResp', result.output.strip())

    @mock.patch('qrl.cli.qrl_pb2_grpc.PublicAPIStub', autospec=True)
    def test_tx_transfer_encrypted_wallet(self, mock_stub):
        tx_pbdata_serialized_to_string = b"\n\x02\\n\x1aC\x01\x02\x00\x80\x9dg/U\xb1N\xf2_\x0e~j%\xb2\x15\x05\xa7y\x19\x8f\xc0>\x05`\x90\xe3>\xaa\x9a(\xd3\xc7U\x91\xbab\x90{\xaa^\xadQ\xca\xbf\xd3\xbc\xd9\x93\xf0:D\xca\xd8v\x97\x08\xa8x\x9c-\n4\xd6e:,\n'\x01\x06\x00\x95O\x16\xafx\xe3\x94Y\rc\x7f\x10D\xcd\x9f\xaf<\xc7\xf4\xa2\x93f\xaa\r\x8c\xa3 \xe40\x0bZ\xfe\xb0xy\xbb\x12\x01\x00"  # noqa
        mock_tx = qrl_pb2.Transaction()
        mock_tx.ParseFromString(tx_pbdata_serialized_to_string)

        m_transferCoinsResp = mock.MagicMock(name='a fake transferCoinsResp')
        m_transferCoinsResp.extended_transaction_unsigned.tx = mock_tx

        m_pushTransactionResp = mock.MagicMock(name='a fake pushTransactionResp', error_code=3)

        attrs = {
            "name": "my fake stub",
            "TransferCoins.return_value": m_transferCoinsResp,
            "PushTransaction.return_value": m_pushTransactionResp
        }
        mock_stub_instance = mock.MagicMock(**attrs)

        mock_stub.name = 'a fake qrl_pb2_grpc.PublicAPIStub'
        mock_stub.return_value = mock_stub_instance

        # Simplest use case
        self.runner.invoke(qrl_cli, ["wallet_encrypt"], input='password\npassword\n')
        result = self.runner.invoke(qrl_cli, ["tx_transfer", "--src=0", "--master=", "--dsts={}".format(qaddr_1),
                                              "--amounts=1", "--fee=0", "--ots_key_index=0"],
                                    input='password\n')
        self.assertEqual(result.exit_code, 0)
        self.assertIn('a fake pushTransactionResp', result.output.strip())

    @mock.patch('qrl.cli.qrl_pb2_grpc.PublicAPIStub', autospec=True)
    def test_tx_transfer_invalid_input(self, mock_stub):
        tx_pbdata_serialized_to_string = b"\n\x02\\n\x1aC\x01\x02\x00\x80\x9dg/U\xb1N\xf2_\x0e~j%\xb2\x15\x05\xa7y\x19\x8f\xc0>\x05`\x90\xe3>\xaa\x9a(\xd3\xc7U\x91\xbab\x90{\xaa^\xadQ\xca\xbf\xd3\xbc\xd9\x93\xf0:D\xca\xd8v\x97\x08\xa8x\x9c-\n4\xd6e:,\n'\x01\x06\x00\x95O\x16\xafx\xe3\x94Y\rc\x7f\x10D\xcd\x9f\xaf<\xc7\xf4\xa2\x93f\xaa\r\x8c\xa3 \xe40\x0bZ\xfe\xb0xy\xbb\x12\x01\x00"  # noqa
        mock_tx = qrl_pb2.Transaction()
        mock_tx.ParseFromString(tx_pbdata_serialized_to_string)

        m_transfer_coins_resp = mock.MagicMock(name='a fake transferCoinsResp')
        m_transfer_coins_resp.extended_transaction_unsigned.tx = mock_tx

        m_push_transaction_resp = mock.MagicMock(name='a fake pushTransactionResp', error_code=3)

        attrs = {
            "name": "my fake stub",
            "TransferCoins.return_value": m_transfer_coins_resp,
            "PushTransaction.return_value": m_push_transaction_resp
        }
        mock_stub_instance = mock.MagicMock(**attrs)

        mock_stub.name = 'a fake qrl_pb2_grpc.PublicAPIStub'
        mock_stub.return_value = mock_stub_instance

        # What if I use a ots_key_index larger than the wallet's tree height?
        result = self.runner.invoke(qrl_cli, [
            "tx_transfer",
            "--src=0",
            "--master=",
            "--dsts={}".format(qaddr_1),
            "--amounts=1",
            "--fee=0",
            "--ots_key_index=16"], input='16')
        self.assertEqual(result.exit_code, 1)
        self.assertIn('OTS key index must be between 0 and 15', result.output.strip())

        # dsts and amounts with different lengths should fail
        dsts = [qaddr_1, qaddr_2, qaddr_3]
        amounts = ["1", "2"]
        result = self.runner.invoke(qrl_cli, [
            "tx_transfer",
            "--src=0",
            "--master=",
            "--dsts={}".format(" ".join(dsts)),
            "--amounts={}".format(" ".join(amounts)),
            "--fee=0",
            "--ots_key_index=0"
        ])
        self.assertEqual(result.exit_code, 1)
        self.assertIn('dsts and amounts should be the same length', result.output.strip())

    @mock.patch('qrl.cli.qrl_pb2_grpc.PublicAPIStub', autospec=True)
    def test_tx_token(self, mock_stub):
        m_push_transaction_resp = mock.MagicMock(name='mock pushTransactionResp')
        m_push_transaction_resp.error_code = 'This was a test'

        attrs = {"name": "this should be qrl_pb2_grpc.PublicAPIStub(channel)",
                 "PushTransaction.return_value": m_push_transaction_resp}
        mock_stub_instance = mock.MagicMock(**attrs)

        mock_stub.name = "this should be qrl_pb2_grpc.PublicAPIStub"
        mock_stub.return_value = mock_stub_instance

        wallet = open_wallet()
        owner_address = wallet["addresses"][0]["address"]
        typed_in_input = '\n'.join([owner_address, '100']) + '\n'
        result = self.runner.invoke(qrl_cli, [
            "tx_token",
            "--src=0",
            "--master=",
            "--symbol=TST",
            "--name=TEST",
            "--owner={}".format(owner_address),
            "--decimals=1",
            "--fee=0",
            "--ots_key_index=0"
        ], input=typed_in_input)
        self.assertIn('This was a test', result.output)

    @mock.patch('qrl.cli.qrl_pb2_grpc.PublicAPIStub', autospec=True)
    def test_tx_token_invalid_input(self, mock_stub):
        m_push_transaction_resp = mock.MagicMock(name='mock pushTransactionResp')
        m_push_transaction_resp.error_code = 'This was a test'

        attrs = {"name": "this should be qrl_pb2_grpc.PublicAPIStub(channel)",
                 "PushTransaction.return_value": m_push_transaction_resp}
        mock_stub_instance = mock.MagicMock(**attrs)

        mock_stub.name = "this should be qrl_pb2_grpc.PublicAPIStub"
        mock_stub.return_value = mock_stub_instance

        wallet = open_wallet()
        owner_address = wallet["addresses"][0]["address"]
        typed_in_input = '\n'.join([owner_address, '100']) + '\n'

        # Weird token names and symbols shouldn't work
        result = self.runner.invoke(qrl_cli,
                                    [
                                        "tx_token",
                                        "--src=0",
                                        "--master=",
                                        "--symbol=thequickbrownfoxjumpsoverthelazydog",
                                        "--name=Seriouslyimgonnastarttalkingandneverendbecausethatsjusthowidothings "
                                        "can i have spaces in here what about |nny characters",
                                        "--owner={}".format(owner_address),
                                        "--decimals=1",
                                        "--fee=0",
                                        "--ots_key_index=0"
                                    ],
                                    input=typed_in_input
                                    )

        print(result.output)
        self.assertEqual(result.exit_code, 1)
        self.assertIn('must be shorter than', result.output.strip())

        # An outrageous decimal precision shouldn't work either
        result = self.runner.invoke(qrl_cli,
                                    [
                                        "tx_token",
                                        "--src=0",
                                        "--master=",
                                        "--symbol=TST",
                                        "--name=TEST",
                                        "--owner={}".format(owner_address),
                                        "--decimals=1000",
                                        "--fee=0",
                                        "--ots_key_index=0"
                                    ],
                                    input=typed_in_input
                                    )

        print(result.output)
        print(result.exc_info)
        self.assertEqual(result.exit_code, 1)

    @mock.patch('qrl.cli.qrl_pb2_grpc.PublicAPIStub', autospec=True)
    def test_tx_transfertoken(self, mock_stub):
        m_push_transaction_resp = mock.MagicMock(name='mock pushTransactionResp')
        m_push_transaction_resp.error_code = 'This was a test'

        attrs = {"name": "this should be qrl_pb2_grpc.PublicAPIStub(channel)",
                 "PushTransaction.return_value": m_push_transaction_resp}
        mock_stub_instance = mock.MagicMock(**attrs)

        mock_stub.name = "this should be qrl_pb2_grpc.PublicAPIStub"
        mock_stub.return_value = mock_stub_instance

        txhash = '267d9c6e192c78b192b6e835411d30f7cb605ffe9632d668489c579e4230f3c6'  # from sample tx_token run

        result = self.runner.invoke(qrl_cli,
                                    [
                                        "tx_transfertoken",
                                        "--src=0",
                                        "--master=",
                                        "--token_txhash={}".format(txhash),
                                        "--dsts={}".format(qaddr_1),
                                        "--amounts=10",
                                        "--decimals=10",
                                        "--fee=0",
                                        "--ots_key_index=0"
                                    ])

        print(result.output)
        self.assertEqual(result.exit_code, 0)
        self.assertEqual(result.output.strip(), 'This was a test')

    @mock.patch('qrl.cli.qrl_pb2_grpc.PublicAPIStub', autospec=True)
    def test_tx_transfertoken_invalid_input(self, mock_stub):
        m_push_transaction_resp = mock.MagicMock(name='mock pushTransactionResp')
        m_push_transaction_resp.error_code = 'This was a test'

        attrs = {"name": "this should be qrl_pb2_grpc.PublicAPIStub(channel)",
                 "PushTransaction.return_value": m_push_transaction_resp}
        mock_stub_instance = mock.MagicMock(**attrs)

        mock_stub.name = "this should be qrl_pb2_grpc.PublicAPIStub"
        mock_stub.return_value = mock_stub_instance

        txhash = '267d9c6e192c78b192b6e835411d30f7cb605ffe9632d668489c579e4230f3c6'  # from sample tx_token run

        # Invalid txhash shouldn't work.
        result = self.runner.invoke(qrl_cli,
                                    [
                                        "tx_transfertoken",
                                        "--src=0",
                                        "--master=",
                                        "--token_txhash={}".format(txhash[3:]),
                                        "--dsts={}".format(qaddr_1),
                                        "--amounts=10",
                                        "--decimals=10",
                                        "--fee=0",
                                        "--ots_key_index=0"],
                                    )

        print(result.output)
        self.assertEqual(result.exit_code, 1)
        self.assertIn('hex string is expected to have an even number of characters', result.output.strip())

        # If decimals is different from the original token_txhash definition, it shouldn't work either.
        # But maybe this is for integration tests.
        # result = self.runner.invoke(qrl_cli,
        #                             ["tx_transfertoken", "--src=0", "--master=",
        #                              "--token_txhash={}".format(txhash), "--dsts={}".format(qaddr_1), "--amounts=10",
        #                              "--decimals=999", "--fee=0", "--ots_key_index=1"],
        #                             )
        # self.assertEqual(result.exit_code, 1)
        # self.assertNotIn('This was a test', result.output.strip())

    @mock.patch('qrl.cli.qrl_pb2_grpc.PublicAPIStub', autospec=True)
    def test_token_list(self, mock_stub):
        m_address_state_resp = mock.MagicMock(name='this should be the addressStateResp')
        m_address_state_resp.state.tokens = {'aabb00': 10, 'ccbb11': 100}
        transaction = mock.MagicMock()
        transaction.token = mock.MagicMock()
        transaction.token.name = b"NAME"
        transaction.token.symbol = b"SYM"
        tx_extended = mock.MagicMock(tx=transaction)
        get_object_resp = mock.MagicMock(transaction=tx_extended)

        mock_stub_instance = mock.MagicMock(name='this should be qrl_pb2_grpc.PublicAPIStub(channel)')
        mock_stub_instance.GetAddressState.return_value = m_address_state_resp
        mock_stub_instance.GetObject.return_value = get_object_resp

        mock_stub.name = 'this should be qrl_pb2_grpc.PublicAPIStub'
        mock_stub.return_value = mock_stub_instance

        result = self.runner.invoke(qrl_cli, ["token_list", "--owner={}".format(qaddr_1)])

        self.assertIn('Hash: {}\nSymbol: {}\nName: {}\nBalance: 10'.format('aabb00',
                                                                           transaction.token.symbol.decode(),
                                                                           transaction.token.name.decode()), result.output)
        self.assertIn('Hash: {}\nSymbol: {}\nName: {}\nBalance: 100'.format('ccbb11',
                                                                            transaction.token.symbol.decode(),
                                                                            transaction.token.name.decode()), result.output)

    @mock.patch('qrl.cli.qrl_pb2_grpc.PublicAPIStub', autospec=True)
    def test_token_list_invalid_input(self, mock_stub):
        m_address_state_resp = mock.MagicMock(name='this should be the addressStateResp')
        m_address_state_resp.state.tokens = {qaddr_1: 10, qaddr_2: 100}

        mock_stub_instance = mock.MagicMock(name='this should be qrl_pb2_grpc.PublicAPIStub(channel)')
        mock_stub_instance.GetAddressState.return_value = m_address_state_resp

        mock_stub.name = 'this should be qrl_pb2_grpc.PublicAPIStub'
        mock_stub.return_value = mock_stub_instance

        # Malformed Qaddress should fail!
        result = self.runner.invoke(qrl_cli, ["token_list", "--owner={}".format(qaddr_1[:-1])])

        self.assertEqual(1, result.exit_code)
        self.assertIn('hex string is expected to have an even number of characters', result.output.strip())



from pyqryptonight.pyqryptonight import PoWHelper

from qrl.core import config


class DifficultyTracker(object):
    def __init__(self):
        pass

    @staticmethod
    def get_target(current_difficulty):
        ph = PoWHelper(kp=config.dev.kp,
                       set_point=config.dev.mining_setpoint_blocktime)
        return ph.getTarget(current_difficulty)

    @staticmethod
    def get(measurement,
            parent_difficulty):

        ph = PoWHelper(kp=config.dev.kp,
                       set_point=config.dev.mining_setpoint_blocktime)

        current_difficulty = ph.getDifficulty(measurement=measurement,
                                              parent_difficulty=parent_difficulty)

        current_target = ph.getTarget(current_difficulty)
        return current_difficulty, current_target



# coding=utf-8
# Distributed under the MIT software license, see the accompanying
# file LICENSE or http://www.opensource.org/licenses/mit-license.php.


class MockFunction:
    def __init__(self, default=None):
        self.data = dict()
        self.default = default

    def put(self, key, value):
        self.data[key] = value

    def get(self, key):
        if key in self.data:
            return self.data[key]

        return self.default

    def remove(self, key):
        del self.data[key]



import time
import multiprocessing
from pyqryptonight.pyqryptonight import Qryptominer


def measure(threads):
    class CustomQMiner(Qryptominer):
        def __init__(self):
            Qryptominer.__init__(self)

        def solutionEvent(self, nonce):
            print("Hey a solution has been found!", nonce)
            self.python_nonce = nonce

    input_bytes = [0x03, 0x05, 0x07, 0x09, 0x03, 0x05, 0x07, 0x09, 0x03, 0x05, 0x07, 0x09, 0x03, 0x05, 0x07, 0x09,
                   0x03, 0x05, 0x07, 0x09, 0x03, 0x05, 0x07, 0x09, 0x03, 0x05, 0x07, 0x09, 0x03, 0x05, 0x07, 0x09,
                   0x03, 0x05, 0x07, 0x09, 0x03, 0x05, 0x07, 0x09, 0x03, 0x05, 0x07]
    target = [
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
    ]

    # Create a customized miner
    qm = CustomQMiner()

    # Set input bytes, nonce
    qm.start(input=input_bytes,
             nonceOffset=0,
             target=target,
             thread_count=threads)

    # Python can sleep or do something else.. the callback will happen in the background
    time.sleep(2)

    return qm.hashRate()


def main():
    cpu_count = multiprocessing.cpu_count()
    for threads in range(1, cpu_count + 1):
        m = []
        for _ in range(3):
            print('.', end='', flush=True)
            m.append(measure(threads))
        print("threads: {:>3}/{:>3}  avg: {:7.2f}H/s    {}]".format(threads, cpu_count, sum(m) / 3, m))


if __name__ == '__main__':
    main()



from unittest import TestCase, expectedFailure
from mock import Mock, patch, PropertyMock, mock

from qrl.core import config
from qrl.core.misc import logger
from qrl.core.AddressState import AddressState
from qrl.core.Block import Block
from qrl.core.BlockMetadata import BlockMetadata
from qrl.core.ESyncState import ESyncState
from qrl.core.qrlnode import QRLNode
from qrl.core.State import State
from qrl.core.ChainManager import ChainManager
from qrl.core.TransactionPool import TransactionPool
from qrl.core.txs.CoinBase import CoinBase
from qrl.core.txs.TransferTransaction import TransferTransaction
from qrl.core.p2p.p2pprotocol import P2PProtocol
from qrl.core.p2p.p2pPeerManager import P2PPeerManager
from qrl.core.p2p.p2pChainManager import P2PChainManager
from qrl.core.node import POW
from qrl.generated import qrl_pb2
from pyqrllib.pyqrllib import hstr2bin

from tests.misc.MockHelper.mock_get_tx_metadata import GetTXMetadata
from tests.core.test_State import gen_blocks
from tests.misc.helper import set_qrl_dir, get_alice_xmss, get_slave_xmss, replacement_getTime, get_random_xmss

logger.initialize_default()


alice = get_alice_xmss()
slave = get_slave_xmss()


@patch('qrl.core.misc.ntp.getTime', new=replacement_getTime)
class TestQRLNodeReal(TestCase):
    @patch('qrl.core.misc.ntp.getTime', new=replacement_getTime)
    def setUp(self):
        # You have to set_qrl_dir to an empty one, otherwise State will have some Transactions from disk
        with set_qrl_dir('no_data'):
            self.db_state = State()
            self.chainmanager = ChainManager(self.db_state)
            self.qrlnode = QRLNode(mining_address=b'')
            self.qrlnode.set_chain_manager(self.chainmanager)

    @patch('qrl.core.qrlnode.QRLNode.block_height', new_callable=PropertyMock, return_value=19)
    def test_get_latest_blocks(self, m_height):
        # [Block 0, Block 1, Block 2... Block 19]
        blocks = gen_blocks(20, self.db_state, alice.address)

        # get_latest_blocks(offset=0, count=1) from [Block 0.... Block 19] should return [Block 19]
        latest_blocks = self.qrlnode.get_latest_blocks(0, 1)
        self.assertTrue(latest_blocks[0] == blocks[19])  # These are different instances, so cannot use assertEqual

        # get_latest_blocks(offset=0, count=2) from [Block 0.... Block 19] should return [Block 18, Block 19]
        latest_blocks = self.qrlnode.get_latest_blocks(0, 2)
        self.assertTrue(latest_blocks[0] == blocks[18])
        self.assertTrue(latest_blocks[1] == blocks[19])

        # get_latest_blocks(offset=10, count=3) from [Block 0.... Block 19] should return [Block 7 - Block 9]
        latest_blocks = self.qrlnode.get_latest_blocks(10, 3)
        self.assertTrue(latest_blocks[0] == blocks[7])
        self.assertTrue(latest_blocks[1] == blocks[8])
        self.assertTrue(latest_blocks[2] == blocks[9])


@patch('qrl.core.misc.ntp.getTime', new=replacement_getTime)
class TestQRLNode(TestCase):
    @patch('qrl.core.misc.ntp.getTime', new=replacement_getTime)
    def setUp(self):
        self.db_state = Mock(autospec=State, name='mocked State')
        # self.db_state = State()

        self.qrlnode = QRLNode(mining_address=b'')

        # As the QRLNode is instantiated and torn down for each test, the minuscule or negative diff between present
        # time and start_time can cause problems.
        self.qrlnode.start_time -= 10

        self.qrlnode._pow = Mock(autospec=POW)
        self.qrlnode._pow.miner.lock.__enter__ = Mock()
        self.qrlnode._pow.miner.lock.__exit__ = Mock()

        self.qrlnode.peer_manager = Mock(autospec=P2PPeerManager, name='mock P2PPeerManager',
                                         known_peer_addresses=set())
        self.qrlnode.p2pchain_manager = Mock(autospec=P2PChainManager, name='mock P2PChainManager')

        self.chain_manager = ChainManager(self.db_state)
        self.chain_manager.tx_pool = Mock(autospec=TransactionPool)
        mock_last_block = Mock(autospec=Block, name='mock last Block', block_number=2, headerhash=b'deadbeef')
        self.chain_manager._last_block = mock_last_block
        self.chain_manager.get_unconfirmed_transaction = Mock()

        self.qrlnode.set_chain_manager(self.chain_manager)

    def test_monitor_chain_state_no_peer_with_higher_difficulty_found(self):
        """
        QRLNode.monitor_chain_state() basically:
        1. Tells P2PPeerManager to clean the list of channels/P2PProtocols, i.e. remove any we haven't heard
        from for a long time, and make sure our list of channels and the list of their statuses is in sync.
        2. Gets the last block from the State
        3. Broadcasts our State based on the last Block to our peers
        4. Ask P2PPeerManager to return the channel (if any) who has a higher difficulty than our chain.
        If no channel with a higher difficulty is found, nothing else happens.
        """
        m_block = Mock(autospec=Block, name='mock last Block', block_number=2, headerhash=b'deadbeef')
        m_block_metadata = Mock(autospec=BlockMetadata, cumulative_difficulty=hstr2bin('01'))
        self.qrlnode.peer_manager.get_better_difficulty.return_value = None
        self.db_state.get_block_metadata.return_value = m_block_metadata
        self.chain_manager.last_block.return_value = m_block

        self.qrlnode.monitor_chain_state()

        self.qrlnode.peer_manager.monitor_chain_state.assert_called_once()
        self.qrlnode.peer_manager.get_better_difficulty.assert_called_once()

    def test_monitor_chain_state_peer_with_higher_difficulty_found(self):
        """
        QRLNode.monitor_chain_state() basically:
        1. Tells P2PPeerManager to clean the list of channels/P2PProtocols, i.e. remove any we haven't heard
        from for a long time, and make sure our list of channels and the list of their statuses is in sync.
        2. Gets the last block from the State
        3. Broadcasts our State based on the last Block to our peers
        4. Ask P2PPeerManager to return the channel (if any) who has a higher difficulty than our chain.
        If a channel with a higher difficulty is found, get its list of headerhashes.
        """
        m_block = Mock(autospec=Block, name='mock last Block', block_number=2, headerhash=b'deadbeef')
        m_block_metadata = Mock(autospec=BlockMetadata, cumulative_difficulty=hstr2bin('01'))
        m_channel = Mock(autospec=P2PProtocol, addr_remote='1.1.1.1')
        self.qrlnode.peer_manager.get_better_difficulty.return_value = m_channel
        self.db_state.get_block_metadata.return_value = m_block_metadata
        self.chain_manager._last_block = m_block

        self.qrlnode.monitor_chain_state()

        self.qrlnode.peer_manager.monitor_chain_state.assert_called_once()
        self.qrlnode.peer_manager.get_better_difficulty.assert_called_once()
        m_channel.send_get_headerhash_list.assert_called_once()

    def test_start_listening(self):
        """
        start_listening() is a convenience function that creates a P2PFactory and connects various components to it
        so that the P2PFactory can function.
        It connects itself, the ChainManager, and the PeerManager to the P2PFactory.
        Then it just tells the P2PFactory to start listening.
        Not much to test here.
        """
        self.assertIsNone(self.qrlnode._p2pfactory)
        self.qrlnode.start_listening()
        self.assertIsNotNone(self.qrlnode._p2pfactory)

    def test_get_address_is_used(self):
        """
        QRLNode.get_address_is_used() asks the DB (State) if a particular address has ever been used.
        It also validates the address before sending it to the State.
        """
        self.db_state.get_address_is_used.return_value = True
        result = self.qrlnode.get_address_is_used(alice.address)
        self.assertTrue(result)

        self.db_state.get_address_is_used.return_value = False
        result = self.qrlnode.get_address_is_used(alice.address)
        self.assertFalse(result)

        with self.assertRaises(ValueError):
            self.qrlnode.get_address_is_used(b'fdsa')

    def test_get_transactions_by_address(self):
        """
        QRLNode.get_transactions_by_address() returns all the changes in balance caused by a transaction.
        """
        mock_get_tx_metadata = GetTXMetadata()

        xmss = get_alice_xmss()
        xmss2 = get_random_xmss()
        addr_state = AddressState.get_default(xmss.address)
        addr_state.pbdata.balance = 100

        tx1 = CoinBase.create(100, xmss.address, 5)
        mock_get_tx_metadata.register_tx_metadata(tx1, 5)
        addr_state.transaction_hashes.append(tx1.txhash)

        tx2 = TransferTransaction.create(addrs_to=[xmss2.address],
                                         amounts=[10],
                                         fee=1,
                                         xmss_pk=xmss.pk)
        tx2.sign(xmss)
        mock_get_tx_metadata.register_tx_metadata(tx2, 99)
        addr_state.transaction_hashes.append(tx2.txhash)

        tx3 = TransferTransaction.create(addrs_to=[xmss.address],
                                         amounts=[100],
                                         fee=1,
                                         xmss_pk=xmss2.pk)
        tx3.sign(xmss)
        mock_get_tx_metadata.register_tx_metadata(tx3, 101)
        addr_state.transaction_hashes.append(tx3.txhash)

        self.db_state.get_address_state.return_value = addr_state
        self.db_state.get_tx_metadata = mock_get_tx_metadata.get_tx_metadata
        result, balance = self.qrlnode.get_transactions_by_address(alice.address)
        self.assertEqual(len(result), 3)

        self.assertEqual(result[0].amount, 100)
        self.assertEqual(result[0].out, False)

        self.assertEqual(result[1].amount, 11)
        self.assertEqual(result[1].out, True)

        self.assertEqual(result[2].amount, 100)
        self.assertEqual(result[2].out, False)

        self.assertEqual(balance, 189)

    def test_get_address_state(self):
        """
        QRLNode.get_address_state() asks the DB (State) for an Address's AddressState, like its nonce, ots index...
        It also validates the address before sending it to the State.
        """
        m_addr_state = Mock(autospec=AddressState)
        self.db_state.get_address_state.return_value = m_addr_state
        result = self.qrlnode.get_address_state(alice.address)
        self.assertEqual(m_addr_state, result)

        # Fetching AddressState for Coinbase Address
        m_addr_state = Mock(autospec=AddressState)
        self.db_state.get_address_state.return_value = m_addr_state
        result = self.qrlnode.get_address_state(config.dev.coinbase_address)
        self.assertEqual(m_addr_state, result)

        with self.assertRaises(ValueError):
            self.qrlnode.get_address_state(b'fdsa')

    def test_get_addr_from(self):
        """
        A master XMSS tree may use a slave XMSS tree to sign for it. If this is the case, we still want to say that the
        TX came from the master XMSS tree, not the slave.
        """
        answer = self.qrlnode.get_addr_from(slave.pk, alice.address)
        self.assertEqual(answer, alice.address)

        answer = self.qrlnode.get_addr_from(slave.pk, None)
        self.assertEqual(answer, slave.address)

    # Just testing that these wrapper functions are doing what they're supposed to do.
    def test_get_transaction(self):
        self.qrlnode.get_transaction(b'a txhash')
        self.db_state.get_tx_metadata.assert_called_once_with(b'a txhash')

    def test_get_unconfirmed_transaction(self):
        self.qrlnode.get_unconfirmed_transaction(b'a txhash')
        self.chain_manager.get_unconfirmed_transaction.assert_called_once_with(b'a txhash')

    def test_get_block_last(self):
        self.assertEqual(self.qrlnode.get_block_last().block_number, 2)

    def test_get_block_from_hash(self):
        self.qrlnode.get_block_from_hash(b'a blockhash')
        self.db_state.get_block.assert_called_once_with(b'a blockhash')

    def test_get_block_from_index(self):
        self.qrlnode.get_block_from_index(3)
        self.db_state.get_block_by_number.assert_called_once_with(3)

    def test_get_blockidx_from_txhash(self):
        self.db_state.get_tx_metadata.return_value = (Mock(name='Mock TX'), 3)
        result = self.qrlnode.get_blockidx_from_txhash(b'a txhash')
        self.assertEqual(result, 3)

        self.db_state.get_tx_metadata.return_value = None
        result = self.qrlnode.get_blockidx_from_txhash(b'a txhash')
        self.assertIsNone(result)

    def test_get_block_to_mine(self):
        m_block = Mock(autospec=Block, name='mock Block')
        m_block_metadata = Mock(autospec=BlockMetadata, name='mock BlockMetadata', block_difficulty=0)

        with mock._patch_object(ChainManager, 'last_block') as m_last_block:
            m_last_block.__get__ = Mock(return_value=m_block)
            self.chain_manager._state.get_block_metadata.return_value = m_block_metadata

            self.qrlnode.get_block_to_mine(alice.address)

            m_last_block.__get__.assert_called_once()
            self.chain_manager._state.get_block_metadata.assert_called_once()
            self.qrlnode._pow.miner.get_block_to_mine.assert_called_once_with(alice.address, self.chain_manager.tx_pool,
                                                                              m_block, 0)

    def test_submit_mined_block(self):
        self.qrlnode.submit_mined_block(b'blob')
        self.qrlnode._pow.miner.submit_mined_block.assert_called_once_with(b'blob')

    def test_get_node_info(self):
        # I guess this test is useful for making sure every part of QRLNode is adequately mocked
        ans = self.qrlnode.get_node_info()
        self.assertIsInstance(ans, qrl_pb2.NodeInfo)

    def test_get_latest_transactions(self):
        """
        This returns the last n txs, just like get_latest_blocks().
        Useful for the Block Explorer, presumably.
        """
        # get_last_txs returns the latest transactions
        self.db_state.get_last_txs.return_value = [Mock(name='mock TX {}'.format(i), i=i) for i in range(19, -1, -1)]

        # Given [0, 1, 2... 19], with offset 0 count 1 should return [19]
        result = self.qrlnode.get_latest_transactions(0, 1)
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0].i, 19)

        # Given [0, 1, 2... 19], with offset 2 count 3 should return [15, 16, 17]
        result = self.qrlnode.get_latest_transactions(2, 3)
        self.assertEqual(len(result), 3)
        self.assertEqual(result[0].i, 17)
        self.assertEqual(result[1].i, 16)
        self.assertEqual(result[2].i, 15)

    @expectedFailure
    def test_get_latest_transactions_unconfirmed(self):
        """
        This should return the last n unconfirmed txs in the txpool.
        Useful for the Block Explorer, presumably.
        FAIL: this returns the first n unconfirmed transactions.
        """
        self.chain_manager.tx_pool.transactions = [(0, Mock(name='mock TX {}'.format(i), i=i)) for i in range(0, 20)]

        # Given [0, 1, 2... 19], with offset 0 count 1 should return [19]
        result = self.qrlnode.get_latest_transactions_unconfirmed(0, 1)
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0].i, 19)

        # Given [0, 1, 2... 19], with offset 2 count 3 should return [15, 16, 17]
        result = self.qrlnode.get_latest_transactions_unconfirmed(2, 3)
        self.assertEqual(len(result), 3)
        self.assertEqual(result[0].i, 15)
        self.assertEqual(result[1].i, 16)
        self.assertEqual(result[2].i, 17)

    def test_get_block_timeseries(self):
        """
        get_block_timeseries() returns a list of stats for the last n blocks.
        The resultant list is in ascending order of time.
        Useful for the Block Explorer, definitely.
        """
        m_blockdps = {
            b'1': Mock(name='BlockDatapoint 1', header_hash=b'1', header_hash_prev=None),
            b'2': Mock(name='BlockDatapoint 2', header_hash=b'2', header_hash_prev=b'1'),
            b'3': Mock(name='BlockDatapoint 3', header_hash=b'3', header_hash_prev=b'2'),
            b'4': Mock(name='BlockDatapoint 4', header_hash=b'4', header_hash_prev=b'3'),
            b'5': Mock(name='BlockDatapoint 5', header_hash=b'5', header_hash_prev=b'4')
        }
        m_blockdps_as_list = [m_blockdps[key] for key in sorted(m_blockdps.keys())]

        self.chain_manager._last_block = Mock(name='Block 5', headerhash=b'5', block_number=5)

        def replacement_get_block_datapoint(headerhash_current):
            return m_blockdps.get(headerhash_current)

        self.chain_manager._state.get_block_datapoint = replacement_get_block_datapoint

        # Get last 5 blocks should return [BlockDatapoint 1, BlockDatapoint 2... BlockDatapoint 5]
        result = self.qrlnode.get_block_timeseries(5)
        result_converted_from_iterator = [r for r in result]
        self.assertEqual(result_converted_from_iterator, m_blockdps_as_list)

        # Get last 3 blocks should return [BlockDatapoint 3, BlockDatapoint 4... BlockDatapoint 5]
        result = self.qrlnode.get_block_timeseries(3)
        result_converted_from_iterator = [r for r in result]
        self.assertEqual(result_converted_from_iterator, m_blockdps_as_list[2:])

        # If we have a blockheight of 0, return []
        with patch('qrl.core.qrlnode.QRLNode.block_height', new_callable=PropertyMock, return_value=0):
            result = self.qrlnode.get_block_timeseries(3)
            result_converted_from_iterator = [r for r in result]  # []
            self.assertFalse(result_converted_from_iterator)

        # If chain_manager.last_block() returns a None, return [] (how is this different from blockheight=0?)
        self.chain_manager._last_block = None
        result = self.qrlnode.get_block_timeseries(5)
        result_converted_from_iterator = [r for r in result]
        self.assertFalse(result_converted_from_iterator)
        self.chain_manager._last_block = Mock(name='Block 5', headerhash=b'5', block_number=5)

        # If we request 6 blocks when we actually have 5, it should still return 5 objects.
        result = self.qrlnode.get_block_timeseries(6)
        result_converted_from_iterator = [r for r in result]
        self.assertEqual(result_converted_from_iterator, m_blockdps_as_list)

    @patch('qrl.core.qrlnode.ChainManager.height', new_callable=PropertyMock, return_value=3)
    def test_get_blockheader_and_metadata(self, m_height):
        blocks = []
        for i in range(0, 4):
            m = Mock(name='mock Block {}'.format(i), i=i)
            m.blockheader.headerhash = str(i).encode()
            blocks.append(m)

        block_metadata = {
            b'0': Mock(name='mock BlockMetadata 0', i=0),
            b'1': Mock(name='mock BlockMetadata 1', i=1),
            b'2': Mock(name='mock BlockMetadata 2', i=2),
            b'3': Mock(name='mock BlockMetadata 3', i=3),
        }

        def replacement_get_block_by_number(idx):
            return blocks[idx]

        def replacement_get_block_metadata(headerhash):
            return block_metadata[headerhash]

        self.db_state.get_block_by_number = replacement_get_block_by_number
        self.db_state.get_block_metadata = replacement_get_block_metadata

        # Because we're just using indexes of a list, we can't actually ever return blocks[0]
        # But this shouldn't be a problem because IRL this uses hashes, not indexes.
        # get_blockheader_and_metadata(0) means get the latest block, which is #3
        result_header, result_metadata = self.qrlnode.get_blockheader_and_metadata(0)
        self.assertEqual(result_header, blocks[-1].blockheader)
        self.assertEqual(result_metadata, block_metadata[b'3'])

        # get block 1 returns the second element of blocks[], which happens to be #1
        result_header, result_metadata = self.qrlnode.get_blockheader_and_metadata(1)
        self.assertEqual(result_header, blocks[1].blockheader)
        self.assertEqual(result_metadata, block_metadata[b'1'])

        # get block 3 returns the fourth element of blocks[], which happens to be #3
        result_header, result_metadata = self.qrlnode.get_blockheader_and_metadata(3)  # get block #3
        self.assertEqual(result_header, blocks[3].blockheader)
        self.assertEqual(result_metadata, block_metadata[b'3'])

        # If get_block_by_number() couldn't find the corresponding block, we should get a (None, None)
        self.db_state.get_block_by_number = Mock(return_value=None)
        result_header, result_metadata = self.qrlnode.get_blockheader_and_metadata(2)
        self.assertIsNone(result_header)
        self.assertIsNone(result_metadata)

    def test_submit_send_tx(self):
        # We verify that P2PFactory.add_unprocessed_txn() won't be called when:
        # TX is None OR
        # pending TXPool is full
        with self.assertRaises(ValueError):
            self.qrlnode.submit_send_tx(None)

        self.chain_manager.tx_pool.is_full_pending_transaction_pool.return_value = True
        m_tx = Mock(name='mock Transaction')
        with self.assertRaises(ValueError):
            self.qrlnode.submit_send_tx(m_tx)

        self.chain_manager.tx_pool.is_full_pending_transaction_pool.return_value = False
        self.qrlnode._p2pfactory = Mock(name='mock P2PFactory')
        self.qrlnode.submit_send_tx(m_tx)
        self.qrlnode._p2pfactory.add_unprocessed_txn.assert_called_once()


@patch('qrl.core.misc.ntp.getTime', new=replacement_getTime)
class TestQRLNodeProperties(TestCase):
    @patch('qrl.core.misc.ntp.getTime', new=replacement_getTime)
    def setUp(self):
        state = Mock()
        self.m_chain_manager = ChainManager(state=state)
        self.m_chain_manager._last_block = None
        self.m_chain_manager._state.get_block_by_number.return_value = None

        self.m_peer_manager = Mock(name='mock P2PPeerManager')

        self.qrlnode = QRLNode(mining_address=b'')
        self.qrlnode.set_chain_manager(self.m_chain_manager)
        self.qrlnode.peer_manager = self.m_peer_manager

    def test_state(self):
        # If qrlnode._p2pfactory is None, then this should be this value
        self.assertEqual(self.qrlnode.state, ESyncState.unknown.value)

        # Else, it should be whatever this part of p2pfactory says
        m_p2pfactory = Mock()
        m_p2pfactory.sync_state.state.value = "test"
        self.qrlnode._p2pfactory = m_p2pfactory
        self.assertEqual(self.qrlnode.state, "test")

    def test_num_connections(self):
        # If qrlnode._p2pfactory is None, then this should return 0
        self.assertEqual(self.qrlnode.num_connections, 0)

        # otherwise it should return what p2pfactory's num_connections says
        m_p2pfactory = Mock(num_connections=5)
        self.qrlnode._p2pfactory = m_p2pfactory
        self.assertEqual(self.qrlnode.num_connections, 5)

    def test_epoch(self):
        self.assertEqual(self.qrlnode.epoch, 0)

        self.m_chain_manager._last_block = Mock(block_number=5)
        self.assertEqual(self.qrlnode.epoch, (5 // config.dev.blocks_per_epoch))

        self.m_chain_manager._last_block = Mock(block_number=256)
        self.assertEqual(self.qrlnode.epoch, (256 // config.dev.blocks_per_epoch))

    def test_uptime_network(self):
        # If there is no block after the genesis block, this property should return 0
        self.assertEqual(self.qrlnode.uptime_network, 0)

        # However, if there is a block after the genesis block, use its timestamp to calculate our uptime.
        with patch('qrl.core.misc.ntp.getTime') as m_getTime:
            self.m_chain_manager._state.get_block_by_number.return_value = Mock(timestamp=1000000)
            m_getTime.return_value = 1500000
            self.assertEqual(self.qrlnode.uptime_network, 500000)

    def test_block_last_reward(self):
        # If last_block() returned None, of course the last reward was 0.
        self.assertEqual(self.qrlnode.block_last_reward, 0)

        # Else it is what the block_reward says it is.
        self.m_chain_manager._last_block = Mock(block_reward=53)
        self.assertEqual(self.qrlnode.block_last_reward, 53)

    def test_block_time_mean(self):
        # FIXME
        # For this function to work, last_block() must not return a None. If it does, bad things will happen.
        self.m_chain_manager._last_block = Mock(name='mock Block')

        # If this particular function returns None, this property should just return the config value
        self.m_chain_manager._state.get_block_metadata.return_value = None
        self.assertEqual(self.qrlnode.block_time_mean, config.dev.mining_setpoint_blocktime)

        # Else, it should consult state.get_measurement()
        self.m_chain_manager._state.get_block_metadata.return_value = Mock(name='mock BlockMetadata')
        self.qrlnode.block_time_mean()
        self.m_chain_manager._state.get_measurement.assert_called_once()

    def test_coin_supply(self):
        with mock._patch_object(ChainManager, 'total_coin_supply') as m_total_coin_supply:
            m_total_coin_supply.__get__ = Mock(return_value=100)
            self.assertEqual(100, self.qrlnode.coin_supply)
            m_total_coin_supply.__get__.assert_called_once()

    def test_coin_supply_max(self):
        # This property should be whatever config says it is.
        self.assertEqual(self.qrlnode.coin_supply_max, config.dev.max_coin_supply)

    def test_get_peers_stat(self):
        self.qrlnode.get_peers_stat()
        self.m_peer_manager.get_peers_stat.assert_called_once()

    def test_connect_peers(self):
        self.qrlnode.connect_peers()
        self.m_peer_manager.connect_peers.assert_called_once()



# coding=utf-8
# Distributed under the MIT software license, see the accompanying
# file LICENSE or http://www.opensource.org/licenses/mit-license.php.
import struct
from queue import PriorityQueue
from typing import Callable, Optional

from google.protobuf.json_format import MessageToJson
from pyqrllib.pyqrllib import bin2hstr  # noqa
from twisted.internet.protocol import Protocol, connectionDone

from qrl.core import config
from qrl.core.OutgoingMessage import OutgoingMessage
from qrl.core.misc import logger, ntp
from qrl.core.p2p.IPMetadata import IPMetadata
from qrl.core.p2p.p2pObservable import P2PObservable
from qrl.generated import qrllegacy_pb2, qrl_pb2

OUT_FACTOR = 0.9
IN_FACTOR = 2.2


# Rename to p2p channel
class P2PProtocol(Protocol):
    def __init__(self):
        self._buffer = bytes()

        # Need to use composition instead of inheritance here
        self._observable = P2PObservable(self)

        self.last_rate_limit_update = 0
        self.rate_limit = config.user.peer_rate_limit
        self.in_counter = 0
        self.out_counter = 0

        self.bytes_sent = 0
        self.outgoing_queue = PriorityQueue(maxsize=config.user.p2p_q_size)

        self._connected_at = ntp.getTime()
        self._valid_message_count = 0

    @property
    def peer(self):
        return IPMetadata(self.transport.getPeer().host, self.transport.getPeer().port)

    @property
    def host(self):
        return IPMetadata(self.transport.getHost().host, self.transport.getHost().port)

    @property
    def connected_at(self):
        return self._connected_at

    @property
    def valid_message_count(self):
        return self._valid_message_count

    @property
    def connection_time(self):
        return ntp.getTime() - self._connected_at

    @property
    def peer_manager(self):
        # FIXME: this is breaking encapsulation
        return self.factory._qrl_node.peer_manager

    @property
    def p2pchain_manager(self):
        # FIXME: this is breaking encapsulation
        return self.factory._qrl_node.p2pchain_manager

    @property
    def tx_manager(self):
        # FIXME: this is breaking encapsulation
        return self.factory._qrl_node.tx_manager

    def register(self, message_type, func: Callable):
        self._observable.register(message_type, func)

    def connectionMade(self):
        if self.factory.add_connection(self):
            self.peer_manager.new_channel(self)
            self.p2pchain_manager.new_channel(self)
            self.tx_manager.new_channel(self)

            self.send_peer_list()
            self.send_version_request()
        else:
            self.loseConnection()

    def connectionLost(self, reason=connectionDone):
        logger.debug('%s disconnected. remainder connected: %d', self.peer, self.factory.num_connections)

        self.factory.remove_connection(self)
        if self.peer_manager:
            self.peer_manager.remove_channel(self)

    def dataReceived(self, data: bytes) -> None:
        self._buffer += data
        total_read = len(self._buffer)

        if total_read > config.dev.max_bytes_out:
            logger.warning('Disconnecting peer %s', self.peer)
            logger.warning('Buffer Size %s', len(self._buffer))
            self.loseConnection()
            return

        read_bytes = [0]

        msg = None
        for msg in self._parse_buffer(read_bytes):
            self.update_counters()
            self.in_counter += 1
            if self.in_counter > self.rate_limit * IN_FACTOR:
                logger.warning("Rate Limit hit by %s %s", self.peer.ip, self.peer.port)
                self.peer_manager.ban_channel(self)
                return

            if self._valid_message_count < config.dev.trust_min_msgcount * 2:
                # Avoid overflows
                self._valid_message_count += 1

            self._observable.notify(msg)

        if msg is not None and read_bytes[0] and msg.func_name != qrllegacy_pb2.LegacyMessage.P2P_ACK:
            p2p_ack = qrl_pb2.P2PAcknowledgement(bytes_processed=read_bytes[0])
            msg = qrllegacy_pb2.LegacyMessage(func_name=qrllegacy_pb2.LegacyMessage.P2P_ACK,
                                              p2pAckData=p2p_ack)
            self.send(msg)

    def update_counters(self):
        time_diff = ntp.getTime() - self.last_rate_limit_update
        if time_diff > 60:
            self.out_counter = 0
            self.in_counter = 0
            self.last_rate_limit_update = ntp.getTime()
            return

    def send_next(self):
        self.update_counters()
        if self.out_counter >= self.rate_limit * OUT_FACTOR:
            return

        if self.bytes_sent < config.dev.max_bytes_out:
            outgoing_bytes = self.get_bytes_from_q()

            if outgoing_bytes:
                self.bytes_sent += len(outgoing_bytes)
                self.transport.write(outgoing_bytes)

    def get_bytes_from_q(self):
        outgoing_bytes = b''
        while not self.outgoing_queue.empty():
            outgoing_msg = self.outgoing_queue.get()[2]
            if not outgoing_msg.is_expired():
                wrapped_message = self._wrap_message(outgoing_msg.message)
                if wrapped_message is not None:
                    if len(wrapped_message) + len(outgoing_bytes) > config.dev.max_bytes_out:
                        self.outgoing_queue.put((outgoing_msg.priority, outgoing_msg.timestamp, outgoing_msg))
                        break
                    outgoing_bytes += wrapped_message

                    self.out_counter += 1
                    if self.out_counter >= self.rate_limit * OUT_FACTOR:
                        break

        return outgoing_bytes

    def send(self, message: qrllegacy_pb2.LegacyMessage):
        priority = self.factory.p2p_msg_priority[message.func_name]
        outgoing_msg = OutgoingMessage(priority, message)
        if self.outgoing_queue.full():
            logger.info("Outgoing Queue Full: Skipping Message Type %s", message.WhichOneof('data'))
            return
        self.outgoing_queue.put((outgoing_msg.priority, outgoing_msg.timestamp, outgoing_msg))
        self.send_next()

    def loseConnection(self):
        self.transport.loseConnection()

    ###################################################
    ###################################################
    ###################################################
    ###################################################
    # Low-level serialization/connections/etc
    # FIXME: This is a temporary refactoring, it will be completely replaced before release

    @staticmethod
    def _wrap_message(protobuf_obj) -> Optional[bytes]:
        """
        Receives a protobuf object and encodes it as (length)(data)
        :return: the encoded message
        :rtype: bytes
        >>> veData = qrllegacy_pb2.VEData(version="version", genesis_prev_hash=b'genesis_hash')
        >>> msg = qrllegacy_pb2.LegacyMessage(func_name=qrllegacy_pb2.LegacyMessage.VE, veData=veData)
        >>> bin2hstr(P2PProtocol._wrap_message(msg))
        '000000191a170a0776657273696f6e120c67656e657369735f68617368'

                msg = qrllegacy_pb2.LegacyMessage(func_name=qrllegacy_pb2.LegacyMessage.PL,
                                          plData=qrllegacy_pb2.PLData(peer_ips=trusted_peers,
                                                                      public_port=config.user.p2p_public_port))

        >>> plData = qrllegacy_pb2.PLData(peer_ips=[])
        >>> msg = qrllegacy_pb2.LegacyMessage(func_name=qrllegacy_pb2.LegacyMessage.PL, plData=plData)
        >>> bin2hstr(P2PProtocol._wrap_message(msg))
        '0000000408012200'
        """
        data = protobuf_obj.SerializeToString()
        if len(data) == 0:
            logger.debug("Skipping message. Zero bytes. %s", MessageToJson(protobuf_obj, sort_keys=True))
            return None

        str_data_len = struct.pack('>L', len(data))
        return str_data_len + data

    def _parse_buffer(self, total_read):
        # FIXME: This parsing/wire protocol needs to be replaced
        """
        >>> from pyqrllib.pyqrllib import hstr2bin
        >>> p=P2PProtocol()
        >>> p._buffer = bytes(hstr2bin('000000191a170a0776657273696f6e120c67656e657369735f68617368'+ \
                                       '000000191a170a0776657273696f6e120c67656e657369735f68617368'))
        >>> messages = p._parse_buffer([0])
        >>> len(list(messages))
        2
        """

        chunk_size = 0

        while self._buffer:
            if len(self._buffer) < 5:
                # Buffer is still incomplete as it doesn't have message size
                return

            ignore_skip = False

            try:
                chunk_size_raw = self._buffer[:4]
                chunk_size = struct.unpack('>L', chunk_size_raw)[0]  # is m length encoded correctly?

                if chunk_size <= 0:
                    logger.debug("<X< %s", bin2hstr(self._buffer))
                    raise Exception("Invalid chunk size <= 0")

                if chunk_size > config.dev.message_buffer_size:
                    raise Exception("Invalid chunk size > message_buffer_size")

                if len(self._buffer) - 4 < chunk_size:  # As 4 bytes includes chunk_size_raw
                    ignore_skip = True  # Buffer is still incomplete as it doesn't have message so skip moving buffer
                    return

                message_raw = self._buffer[4:4 + chunk_size]
                message = qrllegacy_pb2.LegacyMessage()
                message.ParseFromString(message_raw)
                yield message

            except Exception as e:  # no qa
                logger.warning("Problem parsing message. Banning+Dropping connection")
                logger.exception(e)
                self.peer_manager.ban_channel(self)

            finally:
                if not ignore_skip:
                    skip = 4 + chunk_size
                    self._buffer = self._buffer[skip:]
                    total_read[0] += skip

    ###################################################
    ###################################################
    ###################################################
    ###################################################

    def send_version_request(self):
        msg = qrllegacy_pb2.LegacyMessage(func_name=qrllegacy_pb2.LegacyMessage.VE,
                                          veData=qrllegacy_pb2.VEData())
        self.send(msg)

    def send_peer_list(self):
        """
        Get Peers
        Sends the peers list.
        :return:
        """
        trusted_peers = self.peer_manager.trusted_addresses
        logger.debug('<<< Sending connected peers to %s [%s]', self.peer, trusted_peers)

        msg = qrllegacy_pb2.LegacyMessage(func_name=qrllegacy_pb2.LegacyMessage.PL,
                                          plData=qrllegacy_pb2.PLData(peer_ips=trusted_peers,
                                                                      public_port=config.user.p2p_public_port))

        self.send(msg)

    def send_sync(self, synced=False):
        state_str = ''
        if synced:
            state_str = 'Synced'

        msg = qrllegacy_pb2.LegacyMessage(func_name=qrllegacy_pb2.LegacyMessage.SYNC,
                                          syncData=qrllegacy_pb2.SYNCData(state=state_str))
        self.send(msg)

    def send_fetch_block(self, block_idx):
        """
        Fetch Block n
        Sends request for the block number n.
        :return:
        """
        logger.info('<<<Fetching block: %s from %s', block_idx, self.peer)
        msg = qrllegacy_pb2.LegacyMessage(func_name=qrllegacy_pb2.LegacyMessage.FB,
                                          fbData=qrllegacy_pb2.FBData(index=block_idx))
        self.send(msg)

    def send_get_headerhash_list(self, current_block_height):
        start_blocknumber = max(0, current_block_height - config.dev.reorg_limit)
        node_header_hash = qrl_pb2.NodeHeaderHash(block_number=start_blocknumber,
                                                  headerhashes=[])
        msg = qrllegacy_pb2.LegacyMessage(func_name=qrllegacy_pb2.LegacyMessage.HEADERHASHES,
                                          nodeHeaderHash=node_header_hash)
        self.send(msg)



from unittest import TestCase

from qrl.core.misc import logger
from qrl.core.TokenList import TokenList

logger.initialize_default()


class TestTokenList(TestCase):
    tx_hashes = [b'hash1', b'hash2', b'hash3']

    def setUp(self):
        self.tokenlist = TokenList.create(self.tx_hashes)

    def test_token_txhash(self):
        self.assertEqual(self.tokenlist.token_txhash, self.tx_hashes)

    def test_update(self):
        tokenlist = TokenList.create(self.tx_hashes)
        tokenlist.update([b'hash4', b'hash5'])
        self.assertEqual(tokenlist.token_txhash, [b'hash1', b'hash2', b'hash3', b'hash4', b'hash5'])

    def test_json(self):
        json_data = self.tokenlist.to_json()
        tokenlist_new = TokenList.from_json(json_data)
        self.assertEqual(tokenlist_new.pbdata, self.tokenlist.pbdata)



# coding=utf-8
# Distributed under the MIT software license, see the accompanying
# file LICENSE or http://www.opensource.org/licenses/mit-license.php.

import sys
import logging
import traceback

import os
from colorlog import ColoredFormatter
from logging.handlers import RotatingFileHandler

from qrl.core import config

LOG_NAME = 'walletd'

LOG_FILENAME_DEFAULT = config.user.walletd_log_path
LOG_MAXBYTES = 100 * 1024 * 1024
LOG_FORMAT_FULL = '%(asctime)s - %(levelname)s -  %(message)s'
LOG_FORMAT_SMALL = '%(asctime)s - %(message)s'

logger = logging.getLogger(LOG_NAME)


# TODO: Use configuration file instead. Example https://fangpenlin.com/posts/2012/08/26/good-logging-practice-in-python/

def initialize_default(force_console_output=False, log_level=logging.DEBUG):
    logging_target = sys.stderr
    if sys.flags.interactive or force_console_output:
        logger.setLevel(log_level)
        logging_target = sys.stdout

    handler = logging.StreamHandler(logging_target)
    handler.setFormatter(logging.Formatter(LOG_FORMAT_FULL, None))
    logger.addHandler(handler)
    return handler


def log_to_file(filename=LOG_FILENAME_DEFAULT):
    dir_path = os.path.dirname(os.path.realpath(filename))
    os.makedirs(dir_path, exist_ok=True)
    handler = RotatingFileHandler(filename,
                                  mode='a',
                                  maxBytes=LOG_MAXBYTES,
                                  backupCount=2,
                                  encoding=None,
                                  delay=0)
    handler.setLevel(logging.DEBUG)
    logger.addHandler(handler)
    return handler


def _unhandled_exception(etype, value, tb):
    tmp = ['Unhandled exception!\n']
    tmp.extend(traceback.format_exception(etype, value, tb))
    logger.fatal(''.join(tmp))


def set_unhandled_exception_handler():
    sys.excepthook = _unhandled_exception


def get_colors(format_string):
    return ColoredFormatter(
        "%(log_color)s" + format_string,
        datefmt=None,
        reset=True,
        log_colors={
            'DEBUG': 'cyan',
            'INFO': 'white',
            'WARNING': 'yellow',
            'ERROR': 'red',
            'CRITICAL': 'red,bg_white',
        },
        secondary_log_colors={},
        style='%'
    )


def set_colors(enable_colors, formatting):
    for h in logger.handlers:
        if enable_colors and isinstance(h, logging.StreamHandler):
            h.setFormatter(get_colors(formatting))
        else:
            h.setFormatter(logging.Formatter(formatting))


def debug(msg, *args, **kwargs):
    logger.debug(repr(msg)[1:-1], *args, **kwargs)


def info(msg, *args, **kwargs):
    try:
        logger.info(repr(msg)[1:-1], *args, **kwargs)
    except Exception:
        raise Exception


def warning(msg, *args, **kwargs):
    logger.warning(repr(msg)[1:-1], *args, **kwargs)


def error(msg, *args, **kwargs):
    logger.error(repr(msg)[1:-1], *args, **kwargs)


def exception(e):
    error_str = traceback.format_exception(None, e, e.__traceback__)
    logger.error(''.join(error_str))


def fatal(msg, *args, **kwargs):
    logger.fatal(repr(msg)[1:-1], *args, **kwargs)



import contextlib
from math import ceil, log

from mock import mock, MagicMock, Mock
from pyqryptonight.pyqryptonight import StringToUInt256

from qrl.core import config
from qrl.core.Block import Block
from qrl.core.ChainManager import ChainManager
from qrl.core.DifficultyTracker import DifficultyTracker
from qrl.core.GenesisBlock import GenesisBlock
from qrl.core.State import State
from qrl.core.txs.SlaveTransaction import SlaveTransaction
from qrl.core.qrlnode import QRLNode
from tests.misc.helper import get_alice_xmss, get_bob_xmss, set_qrl_dir


class MockedBlockchain(object):
    MAXNUMBLOCKS = 1000

    def __init__(self, qrlnode, time_mock, ntp_mock):
        required_height = ceil(log(self.MAXNUMBLOCKS, 2))
        required_height = int(required_height + required_height % 2)

        self.qrlnode = qrlnode
        self.time_mock = time_mock
        self.ntp_mock = ntp_mock
        self.alice_xmss = get_alice_xmss(xmss_height=required_height)
        self.bob_xmss = get_bob_xmss()

    def create_block(self, prev_hash, mining_address=None):
        if not mining_address:
            mining_address = self.alice_xmss.address
        transactions = []
        block_prev = self.qrlnode.get_block_from_hash(prev_hash)
        block_idx = block_prev.block_number + 1

        if block_idx == 1:
            slave_tx = SlaveTransaction.create(slave_pks=[self.bob_xmss.pk],
                                               access_types=[0],
                                               fee=0,
                                               xmss_pk=self.alice_xmss.pk)
            slave_tx.sign(self.alice_xmss)
            slave_tx._data.nonce = 1
            transactions = [slave_tx]

        time_offset = 60
        if block_idx % 2 == 0:
            time_offset += 2

        self.time_mock.return_value = self.time_mock.return_value + time_offset
        self.ntp_mock.return_value = self.ntp_mock.return_value + time_offset

        block_new = Block.create(block_number=block_idx,
                                 prev_headerhash=block_prev.headerhash,
                                 prev_timestamp=block_prev.timestamp,
                                 transactions=transactions,
                                 miner_address=mining_address)

        while not self.qrlnode._chain_manager.validate_mining_nonce(blockheader=block_new.blockheader):
            block_new.set_nonces(block_new.mining_nonce + 1, 0)

        return block_new

    def validate(self, block):
        if not block.validate(self.qrlnode._chain_manager, {}):
            raise Exception('Block Validation Failed')

        return True

    def add_block(self, block):
        self.validate(block)
        return self.qrlnode._chain_manager.add_block(block)

    def add_new_block(self, mining_address=None):
        block_prev = self.qrlnode.get_block_last()
        block_new = self.create_block(prev_hash=block_prev.headerhash, mining_address=mining_address)
        self.add_block(block_new)

    @staticmethod
    @contextlib.contextmanager
    def create(num_blocks, mining_address=None):
        tmp_gen = GenesisBlock()
        start_time = tmp_gen.timestamp + config.dev.mining_setpoint_blocktime
        with mock.patch('qrl.core.misc.ntp.getTime') as ntp_mock, \
                set_qrl_dir('no_data'), \
                State() as state, \
                mock.patch('time.time') as time_mock:  # noqa
            time_mock.return_value = start_time
            ntp_mock.return_value = start_time

            state.get_measurement = MagicMock(return_value=10000000)

            genesis_difficulty = config.user.genesis_difficulty
            try:
                config.user.genesis_difficulty = 10
                genesis_block = GenesisBlock()
                chain_manager = ChainManager(state)
                chain_manager.load(genesis_block)

                chain_manager._difficulty_tracker = Mock()
                dt = DifficultyTracker()
                tmp_difficulty = StringToUInt256('2')
                tmp_target = dt.get_target(tmp_difficulty)

                chain_manager._difficulty_tracker.get = MagicMock(return_value=(tmp_difficulty, tmp_target))

                qrlnode = QRLNode(mining_address=b'')
                qrlnode.set_chain_manager(chain_manager)

                mock_blockchain = MockedBlockchain(qrlnode, time_mock, ntp_mock)
                for block_idx in range(1, num_blocks + 1):
                    mock_blockchain.add_new_block(mining_address)

                yield mock_blockchain
            finally:
                config.user.genesis_difficulty = genesis_difficulty



# coding=utf-8
# Distributed under the MIT software license, see the accompanying
# file LICENSE or http://www.opensource.org/licenses/mit-license.php.

import heapq

from pyqrllib.pyqrllib import bin2hstr

from qrl.core import config
from qrl.core.misc import logger
from qrl.core.Block import Block
from qrl.core.txs.Transaction import Transaction
from qrl.core.txs.CoinBase import CoinBase
from qrl.core.TransactionInfo import TransactionInfo


class TransactionPool:
    # FIXME: Remove tx pool from all method names
    def __init__(self, broadcast_tx):
        self.pending_tx_pool = []
        self.pending_tx_pool_hash = set()
        self.transaction_pool = []  # FIXME: Everyone is touching this
        self.broadcast_tx = broadcast_tx

    @property
    def transactions(self):
        return heapq.nlargest(len(self.transaction_pool), self.transaction_pool)

    def set_broadcast_tx(self, broadcast_tx):
        self.broadcast_tx = broadcast_tx

    def get_pending_transaction(self):
        if len(self.pending_tx_pool_hash) == 0:
            return None
        pending_tx_set = heapq.heappop(self.pending_tx_pool)
        pending_tx = pending_tx_set[1].transaction
        timestamp = pending_tx_set[1].timestamp
        self.pending_tx_pool_hash.remove(pending_tx.txhash)

        return pending_tx, timestamp

    def is_full_pending_transaction_pool(self, ignore_reserve=True) -> bool:
        max_pool_size = config.user.pending_transaction_pool_size

        if ignore_reserve:
            max_pool_size = max_pool_size - config.user.pending_transaction_pool_reserve

        if len(self.pending_tx_pool) >= max_pool_size:
            return True

        return False

    def is_full_transaction_pool(self) -> bool:
        if len(self.transaction_pool) >= config.user.transaction_pool_size:
            return True

        return False

    def update_pending_tx_pool(self, tx, ip, ignore_reserve=True) -> bool:
        if self.is_full_pending_transaction_pool(ignore_reserve):
            return False

        idx = self.get_tx_index_from_pool(tx.txhash)
        if idx > -1:
            return False

        if isinstance(tx, CoinBase):
            logger.warning('Rejected CoinBase Transaction as received without block')
            return False

        if tx.txhash in self.pending_tx_pool_hash:
            return False

        # Since its a min heap giving priority to lower number
        # So -1 multiplied to give higher priority to higher txn
        heapq.heappush(self.pending_tx_pool, [tx.fee * -1, TransactionInfo(tx, -1), ip])
        self.pending_tx_pool_hash.add(tx.txhash)

        return True

    def add_tx_to_pool(self, tx_class_obj, block_number, timestamp: int=None) -> bool:
        if self.is_full_transaction_pool():
            return False

        heapq.heappush(self.transaction_pool, [tx_class_obj.fee, TransactionInfo(tx_class_obj,
                                                                                 block_number,
                                                                                 timestamp)])
        return True

    def get_tx_index_from_pool(self, txhash):
        for i in range(len(self.transaction_pool)):
            txn = self.transaction_pool[i][1].transaction
            if txhash == txn.txhash:
                return i

        return -1

    def remove_tx_from_pool(self, tx: Transaction):
        idx = self.get_tx_index_from_pool(tx.txhash)
        if idx > -1:
            del self.transaction_pool[idx]
            heapq.heapify(self.transaction_pool)

    def remove_tx_in_block_from_pool(self, block_obj: Block):
        for protobuf_tx in block_obj.transactions[1:]:  # Ignore first transaction, as it is a coinbase txn
            tx = Transaction.from_pbdata(protobuf_tx)
            if tx.ots_key < config.dev.max_ots_tracking_index:
                idx = self.get_tx_index_from_pool(tx.txhash)
                if idx > -1:
                    del self.transaction_pool[idx]
            else:
                i = 0
                while i < len(self.transaction_pool):
                    txn = self.transaction_pool[i][1].transaction
                    if txn.PK == tx.PK:
                        if txn.ots_key >= config.dev.max_ots_tracking_index:
                            if txn.ots_key <= tx.ots_key:
                                del self.transaction_pool[i]
                                continue
                    i += 1

        heapq.heapify(self.transaction_pool)

    def add_tx_from_block_to_pool(self, block: Block, current_block_number):
        """
        Move all transactions from block to transaction pool.
        :param block:
        :return:
        """
        for protobuf_tx in block.transactions[1:]:
            if not self.add_tx_to_pool(Transaction.from_pbdata(protobuf_tx), current_block_number):
                logger.warning('Failed to Add transaction into transaction pool')
                logger.warning('Block #%s %s', block.block_number, bin2hstr(block.headerhash))
                return

    def check_stale_txn(self, state, current_block_number):
        i = 0
        while i < len(self.transaction_pool):
            tx_info = self.transaction_pool[i][1]
            if tx_info.is_stale(current_block_number):
                if not tx_info.validate(state):
                    logger.warning('Txn validation failed for tx in tx_pool')
                    self.remove_tx_from_pool(tx_info.transaction)
                    continue

                tx_info.update_block_number(current_block_number)
                self.broadcast_tx(tx_info.transaction)
            i += 1
